# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""
Onnxscript version of "hf-internal-testing_tiny-random-bart".

See: https://huggingface.co/hf-internal-testing/tiny-random-bart
"""

import numpy as np

import onnxscript.ir as ir
from onnxscript import script
from onnxscript.onnx_opset import opset20
from onnxscript.onnx_types import FLOAT, INT64


def make_model(
    encoder_embed_tokens_weight,
    encoder_embed_positions_weight,
    encoder_layers_0_self_attn_k_proj_bias,
    encoder_layers_0_self_attn_layer_norm_weight,
    encoder_layers_0_fc1_bias,
    matmul_257,
    matmul_267,
    matmul_268,
    matmul_270,
    matmul_271,
    matmul_272,
    matmul_273,
    matmul_283,
    matmul_284,
    matmul_286,
    matmul_287,
    matmul_288,
):
    @script()
    def main_graph(input_ids: INT64[1, None]) -> FLOAT[None, None, 16]:
        encoder_layernorm_embedding_bias = opset20.Identity(
            encoder_layers_0_self_attn_layer_norm_weight
        )
        encoder_layernorm_embedding_weight = opset20.Identity(
            encoder_layers_0_self_attn_layer_norm_weight
        )

        encoder_layers_1_final_layer_norm_bias = opset20.Identity(
            encoder_layers_0_self_attn_k_proj_bias
        )
        encoder_layers_1_final_layer_norm_weight = opset20.Identity(
            encoder_layers_0_self_attn_layer_norm_weight
        )

        encoder_layers_1_fc2_bias = opset20.Identity(encoder_layers_0_self_attn_k_proj_bias)
        encoder_layers_1_self_attn_layer_norm_bias = opset20.Identity(
            encoder_layers_0_self_attn_k_proj_bias
        )
        encoder_layers_1_self_attn_layer_norm_weight = opset20.Identity(
            encoder_layers_0_self_attn_layer_norm_weight
        )
        encoder_layers_1_self_attn_out_proj_bias = opset20.Identity(
            encoder_layers_0_self_attn_k_proj_bias
        )
        encoder_layers_1_self_attn_q_proj_bias = opset20.Identity(
            encoder_layers_0_self_attn_k_proj_bias
        )
        encoder_layers_1_self_attn_v_proj_bias = opset20.Identity(
            encoder_layers_0_self_attn_k_proj_bias
        )
        encoder_layers_1_self_attn_k_proj_bias = opset20.Identity(
            encoder_layers_0_self_attn_k_proj_bias
        )
        encoder_layers_0_final_layer_norm_bias = opset20.Identity(
            encoder_layers_0_self_attn_k_proj_bias
        )
        encoder_layers_0_final_layer_norm_weight = opset20.Identity(
            encoder_layers_0_self_attn_layer_norm_weight
        )
        encoder_layers_0_fc2_bias = opset20.Identity(encoder_layers_0_self_attn_k_proj_bias)
        encoder_layers_1_fc1_bias = opset20.Identity(encoder_layers_0_fc1_bias)
        encoder_layers_0_self_attn_out_proj_bias = opset20.Identity(
            encoder_layers_0_self_attn_k_proj_bias
        )
        encoder_layers_0_self_attn_q_proj_bias = opset20.Identity(
            encoder_layers_0_self_attn_k_proj_bias
        )
        encoder_layers_0_self_attn_v_proj_bias = opset20.Identity(
            encoder_layers_0_self_attn_k_proj_bias
        )

        encoder_shape_output_0 = opset20.Shape(input_ids)
        encoder_constant_output_0 = opset20.Constant(value=1)
        encoder_gather_output_0 = opset20.Gather(
            encoder_shape_output_0, encoder_constant_output_0
        )

        encoder_constant_1_output_0 = opset20.Constant(value=[-1])
        unsqueeze_43 = opset20.Constant(value=[0])
        encoder_unsqueeze_output_0 = opset20.Unsqueeze(encoder_gather_output_0, unsqueeze_43)
        encoder_concat_output_0 = opset20.Concat(
            encoder_constant_1_output_0, encoder_unsqueeze_output_0, axis=0
        )
        encoder_reshape_output_0 = opset20.Reshape(
            input_ids, encoder_concat_output_0, allowzero=0
        )
        encoder_embed_tokens_gather_output_0 = opset20.Gather(
            encoder_embed_tokens_weight, encoder_reshape_output_0
        )
        encoder_embed_tokens_constant_output_0 = opset20.Constant(value=[1.0])
        encoder_embed_tokens_mul_output_0 = opset20.Mul(
            encoder_embed_tokens_gather_output_0, encoder_embed_tokens_constant_output_0
        )
        encoder_embed_positions_shape_output_0 = opset20.Shape(input_ids)
        encoder_embed_positions_constant_output_0 = opset20.Constant(value=0)
        encoder_embed_positions_gather_output_0 = opset20.Gather(
            encoder_embed_positions_shape_output_0,
            encoder_embed_positions_constant_output_0,
            axis=0,
        )
        encoder_embed_positions_constant_1_output_0 = opset20.Constant(value=0)
        encoder_embed_positions_cast_output_0 = opset20.Cast(encoder_gather_output_0, to=7)
        encoder_embed_positions_constant_2_output_0 = opset20.Constant(value=1)
        encoder_embed_positions_range_output_0 = opset20.Range(
            encoder_embed_positions_constant_1_output_0,
            encoder_embed_positions_cast_output_0,
            encoder_embed_positions_constant_2_output_0,
        )
        encoder_embed_positions_constant_3_output_0 = opset20.Constant(value=[0])
        encoder_embed_positions_unsqueeze_output_0 = opset20.Unsqueeze(
            encoder_embed_positions_gather_output_0,
            encoder_embed_positions_constant_3_output_0,
        )
        encoder_embed_positions_constant_4_output_0 = opset20.Constant(value=[-1])
        encoder_embed_positions_concat_output_0 = opset20.Concat(
            encoder_embed_positions_unsqueeze_output_0,
            encoder_embed_positions_constant_4_output_0,
            axis=0,
        )
        encoder_embed_positions_constant_5_output_0 = opset20.Constant(value=[-1])
        encoder_embed_positions_reshape_output_0 = opset20.Reshape(
            encoder_embed_positions_concat_output_0,
            encoder_embed_positions_constant_5_output_0,
        )
        encoder_embed_positions_shape_1_output_0 = opset20.Shape(
            encoder_embed_positions_reshape_output_0
        )
        encoder_embed_positions_constantofshape_output_0 = opset20.ConstantOfShape(
            encoder_embed_positions_shape_1_output_0,
            value=ir.tensor(np.array([1], dtype=np.int64)),
        )
        encoder_embed_positions_constant_6_output_0 = opset20.Constant(value=[-1])
        encoder_embed_positions_mul_output_0 = opset20.Mul(
            encoder_embed_positions_constantofshape_output_0,
            encoder_embed_positions_constant_6_output_0,
        )
        encoder_embed_positions_equal_output_0 = opset20.Equal(
            encoder_embed_positions_reshape_output_0, encoder_embed_positions_mul_output_0
        )
        encoder_embed_positions_where_output_0 = opset20.Where(
            encoder_embed_positions_equal_output_0,
            encoder_embed_positions_constantofshape_output_0,
            encoder_embed_positions_reshape_output_0,
        )
        encoder_embed_positions_expand_output_0 = opset20.Expand(
            encoder_embed_positions_range_output_0, encoder_embed_positions_where_output_0
        )
        encoder_embed_positions_constant_7_output_0 = opset20.Constant(value=2)
        encoder_embed_positions_add_output_0 = opset20.Add(
            encoder_embed_positions_expand_output_0,
            encoder_embed_positions_constant_7_output_0,
        )
        encoder_embed_positions_gather_1_output_0 = opset20.Gather(
            encoder_embed_positions_weight, encoder_embed_positions_add_output_0
        )
        encoder_cast_output_0 = opset20.Cast(encoder_embed_positions_gather_1_output_0, to=1)
        encoder_add_output_0 = opset20.Add(
            encoder_embed_tokens_mul_output_0, encoder_cast_output_0
        )
        encoder_layernorm_embedding_layernormalization_output_0 = opset20.LayerNormalization(
            encoder_add_output_0,
            encoder_layernorm_embedding_weight,
            encoder_layernorm_embedding_bias,
            axis=-1,
            epsilon=9.999999747378752e-06,
        )
        encoder_layers_0_self_attn_shape_output_0 = opset20.Shape(
            encoder_layernorm_embedding_layernormalization_output_0
        )
        encoder_layers_0_self_attn_constant_output_0 = opset20.Constant(value=0)
        encoder_layers_0_self_attn_gather_output_0 = opset20.Gather(
            encoder_layers_0_self_attn_shape_output_0,
            encoder_layers_0_self_attn_constant_output_0,
            axis=0,
        )
        encoder_layers_0_self_attn_shape_1_output_0 = opset20.Shape(
            encoder_layernorm_embedding_layernormalization_output_0
        )
        encoder_layers_0_self_attn_constant_1_output_0 = opset20.Constant(value=1)
        encoder_layers_0_self_attn_gather_1_output_0 = opset20.Gather(
            encoder_layers_0_self_attn_shape_1_output_0,
            encoder_layers_0_self_attn_constant_1_output_0,
            axis=0,
        )
        encoder_layers_0_self_attn_q_proj_matmul_output_0 = opset20.MatMul(
            encoder_layernorm_embedding_layernormalization_output_0, matmul_257
        )
        encoder_layers_0_self_attn_q_proj_add_output_0 = opset20.Add(
            encoder_layers_0_self_attn_q_proj_bias,
            encoder_layers_0_self_attn_q_proj_matmul_output_0,
        )
        unsqueeze_88 = opset20.Constant(value=[0])
        encoder_layers_0_self_attn_unsqueeze_output_0 = opset20.Unsqueeze(
            encoder_layers_0_self_attn_gather_output_0, unsqueeze_88
        )
        encoder_layers_0_self_attn_constant_2_output_0 = opset20.Constant(value=[-1])
        encoder_layers_0_self_attn_constant_3_output_0 = opset20.Constant(value=[4])
        encoder_layers_0_self_attn_constant_4_output_0 = opset20.Constant(value=[4])
        encoder_layers_0_self_attn_concat_output_0 = opset20.Concat(
            encoder_layers_0_self_attn_unsqueeze_output_0,
            encoder_layers_0_self_attn_constant_2_output_0,
            encoder_layers_0_self_attn_constant_3_output_0,
            encoder_layers_0_self_attn_constant_4_output_0,
            axis=0,
        )
        unsqueeze_97 = opset20.Constant(value=[0])
        encoder_layers_0_self_attn_unsqueeze_1_output_0 = opset20.Unsqueeze(
            encoder_layers_0_self_attn_gather_output_0, unsqueeze_97
        )
        encoder_layers_0_self_attn_constant_5_output_0 = opset20.Constant(value=[-1])
        encoder_layers_0_self_attn_constant_6_output_0 = opset20.Constant(value=[4])
        encoder_layers_0_self_attn_constant_7_output_0 = opset20.Constant(value=[4])
        encoder_layers_0_self_attn_concat_1_output_0 = opset20.Concat(
            encoder_layers_0_self_attn_unsqueeze_1_output_0,
            encoder_layers_0_self_attn_constant_5_output_0,
            encoder_layers_0_self_attn_constant_6_output_0,
            encoder_layers_0_self_attn_constant_7_output_0,
            axis=0,
        )
        unsqueeze_106 = opset20.Constant(value=[0])
        encoder_layers_0_self_attn_unsqueeze_2_output_0 = opset20.Unsqueeze(
            encoder_layers_0_self_attn_gather_output_0, unsqueeze_106
        )
        encoder_layers_0_self_attn_constant_8_output_0 = opset20.Constant(value=[-1])
        encoder_layers_0_self_attn_constant_9_output_0 = opset20.Constant(value=[4])
        encoder_layers_0_self_attn_constant_10_output_0 = opset20.Constant(value=[4])
        encoder_layers_0_self_attn_concat_2_output_0 = opset20.Concat(
            encoder_layers_0_self_attn_unsqueeze_2_output_0,
            encoder_layers_0_self_attn_constant_8_output_0,
            encoder_layers_0_self_attn_constant_9_output_0,
            encoder_layers_0_self_attn_constant_10_output_0,
            axis=0,
        )

        encoder_layers_0_self_attn_reshape_output_0 = opset20.Reshape(
            encoder_layers_0_self_attn_q_proj_add_output_0,
            encoder_layers_0_self_attn_concat_output_0,
            allowzero=0,
        )
        encoder_layers_0_self_attn_transpose_output_0 = opset20.Transpose(
            encoder_layers_0_self_attn_reshape_output_0, perm=[0, 2, 1, 3]
        )
        encoder_layers_0_self_attn_k_proj_matmul_output_0 = opset20.MatMul(
            encoder_layernorm_embedding_layernormalization_output_0, matmul_267
        )
        encoder_layers_0_self_attn_k_proj_add_output_0 = opset20.Add(
            encoder_layers_0_self_attn_k_proj_bias,
            encoder_layers_0_self_attn_k_proj_matmul_output_0,
        )
        encoder_layers_0_self_attn_v_proj_matmul_output_0 = opset20.MatMul(
            encoder_layernorm_embedding_layernormalization_output_0, matmul_268
        )
        encoder_layers_0_self_attn_v_proj_add_output_0 = opset20.Add(
            encoder_layers_0_self_attn_v_proj_bias,
            encoder_layers_0_self_attn_v_proj_matmul_output_0,
        )
        encoder_layers_0_self_attn_reshape_1_output_0 = opset20.Reshape(
            encoder_layers_0_self_attn_k_proj_add_output_0,
            encoder_layers_0_self_attn_concat_1_output_0,
            allowzero=0,
        )
        encoder_layers_0_self_attn_reshape_2_output_0 = opset20.Reshape(
            encoder_layers_0_self_attn_v_proj_add_output_0,
            encoder_layers_0_self_attn_concat_2_output_0,
            allowzero=0,
        )
        encoder_layers_0_self_attn_transpose_1_output_0 = opset20.Transpose(
            encoder_layers_0_self_attn_reshape_2_output_0, perm=[0, 2, 1, 3]
        )
        encoder_layers_0_self_attn_shape_2_output_0 = opset20.Shape(
            encoder_layers_0_self_attn_transpose_output_0
        )
        encoder_layers_0_self_attn_constant_11_output_0 = opset20.Constant(value=[-1])
        encoder_layers_0_self_attn_constant_12_output_0 = opset20.Constant(
            value=[9223372036854775807]
        )
        encoder_layers_0_self_attn_slice_output_0 = opset20.Slice(
            encoder_layers_0_self_attn_shape_2_output_0,
            encoder_layers_0_self_attn_constant_11_output_0,
            encoder_layers_0_self_attn_constant_12_output_0,
        )
        encoder_layers_0_self_attn_cast_output_0 = opset20.Cast(
            encoder_layers_0_self_attn_slice_output_0, to=1
        )
        encoder_layers_0_self_attn_sqrt_output_0 = opset20.Sqrt(
            encoder_layers_0_self_attn_cast_output_0
        )
        encoder_layers_0_self_attn_constant_13_output_0 = opset20.Constant(value=[1.0])
        encoder_layers_0_self_attn_div_output_0 = opset20.Div(
            encoder_layers_0_self_attn_constant_13_output_0,
            encoder_layers_0_self_attn_sqrt_output_0,
        )
        encoder_layers_0_self_attn_cast_1_output_0 = opset20.Cast(
            encoder_layers_0_self_attn_div_output_0, to=1
        )
        encoder_layers_0_self_attn_transpose_2_output_0 = opset20.Transpose(
            encoder_layers_0_self_attn_reshape_1_output_0, perm=[0, 2, 3, 1]
        )
        encoder_layers_0_self_attn_sqrt_1_output_0 = opset20.Sqrt(
            encoder_layers_0_self_attn_cast_1_output_0
        )
        encoder_layers_0_self_attn_mul_output_0 = opset20.Mul(
            encoder_layers_0_self_attn_transpose_output_0,
            encoder_layers_0_self_attn_sqrt_1_output_0,
        )
        encoder_layers_0_self_attn_sqrt_2_output_0 = opset20.Sqrt(
            encoder_layers_0_self_attn_cast_1_output_0
        )
        encoder_layers_0_self_attn_mul_1_output_0 = opset20.Mul(
            encoder_layers_0_self_attn_transpose_2_output_0,
            encoder_layers_0_self_attn_sqrt_2_output_0,
        )
        encoder_layers_0_self_attn_matmul_output_0 = opset20.MatMul(
            encoder_layers_0_self_attn_mul_output_0, encoder_layers_0_self_attn_mul_1_output_0
        )
        encoder_layers_0_self_attn_softmax_output_0 = opset20.Softmax(
            encoder_layers_0_self_attn_matmul_output_0, axis=-1
        )
        encoder_layers_0_self_attn_matmul_1_output_0 = opset20.MatMul(
            encoder_layers_0_self_attn_softmax_output_0,
            encoder_layers_0_self_attn_transpose_1_output_0,
        )
        encoder_layers_0_self_attn_transpose_3_output_0 = opset20.Transpose(
            encoder_layers_0_self_attn_matmul_1_output_0, perm=[0, 2, 1, 3]
        )
        unsqueeze_145 = opset20.Constant(value=[0])
        encoder_layers_0_self_attn_unsqueeze_3_output_0 = opset20.Unsqueeze(
            encoder_layers_0_self_attn_gather_output_0, unsqueeze_145
        )
        unsqueeze_147 = opset20.Constant(value=[0])
        encoder_layers_0_self_attn_unsqueeze_4_output_0 = opset20.Unsqueeze(
            encoder_layers_0_self_attn_gather_1_output_0, unsqueeze_147
        )
        encoder_layers_0_self_attn_constant_14_output_0 = opset20.Constant(value=[16])
        encoder_layers_0_self_attn_concat_3_output_0 = opset20.Concat(
            encoder_layers_0_self_attn_unsqueeze_3_output_0,
            encoder_layers_0_self_attn_unsqueeze_4_output_0,
            encoder_layers_0_self_attn_constant_14_output_0,
            axis=0,
        )
        encoder_layers_0_self_attn_reshape_3_output_0 = opset20.Reshape(
            encoder_layers_0_self_attn_transpose_3_output_0,
            encoder_layers_0_self_attn_concat_3_output_0,
            allowzero=0,
        )
        encoder_layers_0_self_attn_out_proj_matmul_output_0 = opset20.MatMul(
            encoder_layers_0_self_attn_reshape_3_output_0, matmul_270
        )
        encoder_layers_0_self_attn_out_proj_add_output_0 = opset20.Add(
            encoder_layers_0_self_attn_out_proj_bias,
            encoder_layers_0_self_attn_out_proj_matmul_output_0,
        )
        encoder_layers_0_add_output_0 = opset20.Add(
            encoder_layernorm_embedding_layernormalization_output_0,
            encoder_layers_0_self_attn_out_proj_add_output_0,
        )
        encoder_layers_0_self_attn_layer_norm_layernormalization_output_0 = (
            opset20.LayerNormalization(
                encoder_layers_0_add_output_0,
                encoder_layers_0_self_attn_layer_norm_weight,
                encoder_layernorm_embedding_bias,
                axis=-1,
                epsilon=9.999999747378752e-0,
            )
        )
        encoder_layers_0_fc1_matmul_output_0 = opset20.MatMul(
            encoder_layers_0_self_attn_layer_norm_layernormalization_output_0, matmul_271
        )
        encoder_layers_0_fc1_add_output_0 = opset20.Add(
            encoder_layers_0_fc1_bias, encoder_layers_0_fc1_matmul_output_0
        )
        encoder_layers_0_activation_fn_gelu_output_0 = opset20.Gelu(
            encoder_layers_0_fc1_add_output_0, approximate="none"
        )
        encoder_layers_0_fc2_matmul_output_0 = opset20.MatMul(
            encoder_layers_0_activation_fn_gelu_output_0, matmul_272
        )
        encoder_layers_0_fc2_add_output_0 = opset20.Add(
            encoder_layers_0_fc2_bias, encoder_layers_0_fc2_matmul_output_0
        )
        encoder_layers_0_add_1_output_0 = opset20.Add(
            encoder_layers_0_self_attn_layer_norm_layernormalization_output_0,
            encoder_layers_0_fc2_add_output_0,
        )
        encoder_layers_0_final_layer_norm_layernormalization_output_0 = (
            opset20.LayerNormalization(
                encoder_layers_0_add_1_output_0,
                encoder_layers_0_final_layer_norm_weight,
                encoder_layers_0_final_layer_norm_bias,
                axis=-1,
                epsilon=9.999999747378752e-06,
            )
        )
        encoder_layers_1_self_attn_shape_output_0 = opset20.Shape(
            encoder_layers_0_final_layer_norm_layernormalization_output_0
        )
        encoder_layers_1_self_attn_constant_output_0 = opset20.Constant(value=0)
        encoder_layers_1_self_attn_gather_output_0 = opset20.Gather(
            encoder_layers_1_self_attn_shape_output_0,
            encoder_layers_1_self_attn_constant_output_0,
            axis=0,
        )
        encoder_layers_1_self_attn_shape_1_output_0 = opset20.Shape(
            encoder_layers_0_final_layer_norm_layernormalization_output_0
        )
        encoder_layers_1_self_attn_constant_1_output_0 = opset20.Constant(value=1)
        encoder_layers_1_self_attn_gather_1_output_0 = opset20.Gather(
            encoder_layers_1_self_attn_shape_1_output_0,
            encoder_layers_1_self_attn_constant_1_output_0,
            axis=0,
        )
        encoder_layers_1_self_attn_q_proj_matmul_output_0 = opset20.MatMul(
            encoder_layers_0_final_layer_norm_layernormalization_output_0, matmul_273
        )
        encoder_layers_1_self_attn_q_proj_add_output_0 = opset20.Add(
            encoder_layers_1_self_attn_q_proj_bias,
            encoder_layers_1_self_attn_q_proj_matmul_output_0,
        )
        unsqueeze_176 = opset20.Constant(value=[0])
        encoder_layers_1_self_attn_unsqueeze_output_0 = opset20.Unsqueeze(
            encoder_layers_1_self_attn_gather_output_0, unsqueeze_176
        )
        encoder_layers_1_self_attn_constant_2_output_0 = opset20.Constant(value=[-1])
        encoder_layers_1_self_attn_constant_3_output_0 = opset20.Constant(value=[4])
        encoder_layers_1_self_attn_constant_4_output_0 = opset20.Constant(value=[4])
        encoder_layers_1_self_attn_concat_output_0 = opset20.Concat(
            encoder_layers_1_self_attn_unsqueeze_output_0,
            encoder_layers_1_self_attn_constant_2_output_0,
            encoder_layers_1_self_attn_constant_3_output_0,
            encoder_layers_1_self_attn_constant_4_output_0,
            axis=0,
        )
        unsqueeze_185 = opset20.Constant(value=[0])
        encoder_layers_1_self_attn_unsqueeze_1_output_0 = opset20.Unsqueeze(
            encoder_layers_1_self_attn_gather_output_0, unsqueeze_185
        )
        encoder_layers_1_self_attn_constant_5_output_0 = opset20.Constant(value=[-1])
        encoder_layers_1_self_attn_constant_6_output_0 = opset20.Constant(value=[4])
        encoder_layers_1_self_attn_constant_7_output_0 = opset20.Constant(value=[4])
        encoder_layers_1_self_attn_concat_1_output_0 = opset20.Concat(
            encoder_layers_1_self_attn_unsqueeze_1_output_0,
            encoder_layers_1_self_attn_constant_5_output_0,
            encoder_layers_1_self_attn_constant_6_output_0,
            encoder_layers_1_self_attn_constant_7_output_0,
            axis=0,
        )
        unsqueeze_194 = opset20.Constant(value=[0])
        encoder_layers_1_self_attn_unsqueeze_2_output_0 = opset20.Unsqueeze(
            encoder_layers_1_self_attn_gather_output_0, unsqueeze_194
        )
        encoder_layers_1_self_attn_constant_8_output_0 = opset20.Constant(value=[-1])
        encoder_layers_1_self_attn_constant_9_output_0 = opset20.Constant(value=[4])
        encoder_layers_1_self_attn_constant_10_output_0 = opset20.Constant(value=[4])
        encoder_layers_1_self_attn_concat_2_output_0 = opset20.Concat(
            encoder_layers_1_self_attn_unsqueeze_2_output_0,
            encoder_layers_1_self_attn_constant_8_output_0,
            encoder_layers_1_self_attn_constant_9_output_0,
            encoder_layers_1_self_attn_constant_10_output_0,
            axis=0,
        )
        encoder_layers_1_self_attn_reshape_output_0 = opset20.Reshape(
            encoder_layers_1_self_attn_q_proj_add_output_0,
            encoder_layers_1_self_attn_concat_output_0,
            allowzero=0,
        )
        encoder_layers_1_self_attn_transpose_output_0 = opset20.Transpose(
            encoder_layers_1_self_attn_reshape_output_0, perm=[0, 2, 1, 3]
        )
        encoder_layers_1_self_attn_k_proj_matmul_output_0 = opset20.MatMul(
            encoder_layers_0_final_layer_norm_layernormalization_output_0, matmul_283
        )
        encoder_layers_1_self_attn_k_proj_add_output_0 = opset20.Add(
            encoder_layers_1_self_attn_k_proj_bias,
            encoder_layers_1_self_attn_k_proj_matmul_output_0,
        )
        encoder_layers_1_self_attn_v_proj_matmul_output_0 = opset20.MatMul(
            encoder_layers_0_final_layer_norm_layernormalization_output_0, matmul_284
        )
        encoder_layers_1_self_attn_v_proj_add_output_0 = opset20.Add(
            encoder_layers_1_self_attn_v_proj_bias,
            encoder_layers_1_self_attn_v_proj_matmul_output_0,
        )
        encoder_layers_1_self_attn_reshape_1_output_0 = opset20.Reshape(
            encoder_layers_1_self_attn_k_proj_add_output_0,
            encoder_layers_1_self_attn_concat_1_output_0,
            allowzero=0,
        )
        encoder_layers_1_self_attn_reshape_2_output_0 = opset20.Reshape(
            encoder_layers_1_self_attn_v_proj_add_output_0,
            encoder_layers_1_self_attn_concat_2_output_0,
            allowzero=0,
        )
        encoder_layers_1_self_attn_transpose_1_output_0 = opset20.Transpose(
            encoder_layers_1_self_attn_reshape_2_output_0, perm=[0, 2, 1, 3]
        )
        encoder_layers_1_self_attn_shape_2_output_0 = opset20.Shape(
            encoder_layers_1_self_attn_transpose_output_0
        )
        encoder_layers_1_self_attn_constant_11_output_0 = opset20.Constant(value=[-1])
        encoder_layers_1_self_attn_constant_12_output_0 = opset20.Constant(
            value=[9223372036854775807]
        )
        encoder_layers_1_self_attn_slice_output_0 = opset20.Slice(
            encoder_layers_1_self_attn_shape_2_output_0,
            encoder_layers_1_self_attn_constant_11_output_0,
            encoder_layers_1_self_attn_constant_12_output_0,
        )
        encoder_layers_1_self_attn_cast_output_0 = opset20.Cast(
            encoder_layers_1_self_attn_slice_output_0, to=1
        )
        encoder_layers_1_self_attn_sqrt_output_0 = opset20.Sqrt(
            encoder_layers_1_self_attn_cast_output_0
        )
        encoder_layers_1_self_attn_constant_13_output_0 = opset20.Constant(value=[1.0])
        encoder_layers_1_self_attn_div_output_0 = opset20.Div(
            encoder_layers_1_self_attn_constant_13_output_0,
            encoder_layers_1_self_attn_sqrt_output_0,
        )
        encoder_layers_1_self_attn_cast_1_output_0 = opset20.Cast(
            encoder_layers_1_self_attn_div_output_0, to=1
        )
        encoder_layers_1_self_attn_transpose_2_output_0 = opset20.Transpose(
            encoder_layers_1_self_attn_reshape_1_output_0, perm=[0, 2, 3, 1]
        )
        encoder_layers_1_self_attn_sqrt_1_output_0 = opset20.Sqrt(
            encoder_layers_1_self_attn_cast_1_output_0
        )
        encoder_layers_1_self_attn_mul_output_0 = opset20.Mul(
            encoder_layers_1_self_attn_transpose_output_0,
            encoder_layers_1_self_attn_sqrt_1_output_0,
        )
        encoder_layers_1_self_attn_sqrt_2_output_0 = opset20.Sqrt(
            encoder_layers_1_self_attn_cast_1_output_0
        )
        encoder_layers_1_self_attn_mul_1_output_0 = opset20.Mul(
            encoder_layers_1_self_attn_transpose_2_output_0,
            encoder_layers_1_self_attn_sqrt_2_output_0,
        )
        encoder_layers_1_self_attn_matmul_output_0 = opset20.MatMul(
            encoder_layers_1_self_attn_mul_output_0, encoder_layers_1_self_attn_mul_1_output_0
        )
        encoder_layers_1_self_attn_softmax_output_0 = opset20.Softmax(
            encoder_layers_1_self_attn_matmul_output_0, axis=-1
        )
        encoder_layers_1_self_attn_matmul_1_output_0 = opset20.MatMul(
            encoder_layers_1_self_attn_softmax_output_0,
            encoder_layers_1_self_attn_transpose_1_output_0,
        )
        encoder_layers_1_self_attn_transpose_3_output_0 = opset20.Transpose(
            encoder_layers_1_self_attn_matmul_1_output_0, perm=[0, 2, 1, 3]
        )
        unsqueeze_232 = opset20.Constant(value=[0])
        encoder_layers_1_self_attn_unsqueeze_3_output_0 = opset20.Unsqueeze(
            encoder_layers_1_self_attn_gather_output_0, unsqueeze_232
        )
        unsqueeze_234 = opset20.Constant(value=[0])
        encoder_layers_1_self_attn_unsqueeze_4_output_0 = opset20.Unsqueeze(
            encoder_layers_1_self_attn_gather_1_output_0, unsqueeze_234
        )
        encoder_layers_1_self_attn_constant_14_output_0 = opset20.Constant(value=[16])

        encoder_layers_1_self_attn_concat_3_output_0 = opset20.Concat(
            encoder_layers_1_self_attn_unsqueeze_3_output_0,
            encoder_layers_1_self_attn_unsqueeze_4_output_0,
            encoder_layers_1_self_attn_constant_14_output_0,
            axis=0,
        )
        encoder_layers_1_self_attn_reshape_3_output_0 = opset20.Reshape(
            encoder_layers_1_self_attn_transpose_3_output_0,
            encoder_layers_1_self_attn_concat_3_output_0,
            allowzero=0,
        )
        encoder_layers_1_self_attn_out_proj_matmul_output_0 = opset20.MatMul(
            encoder_layers_1_self_attn_reshape_3_output_0, matmul_286
        )
        encoder_layers_1_self_attn_out_proj_add_output_0 = opset20.Add(
            encoder_layers_1_self_attn_out_proj_bias,
            encoder_layers_1_self_attn_out_proj_matmul_output_0,
        )
        encoder_layers_1_add_output_0 = opset20.Add(
            encoder_layers_0_final_layer_norm_layernormalization_output_0,
            encoder_layers_1_self_attn_out_proj_add_output_0,
        )
        encoder_layers_1_self_attn_layer_norm_layernormalization_output_0 = (
            opset20.LayerNormalization(
                encoder_layers_1_add_output_0,
                encoder_layers_1_self_attn_layer_norm_weight,
                encoder_layers_1_self_attn_layer_norm_bias,
                axis=-1,
                epsilon=9.999999747378752e-06,
            )
        )
        encoder_layers_1_fc1_matmul_output_0 = opset20.MatMul(
            encoder_layers_1_self_attn_layer_norm_layernormalization_output_0, matmul_287
        )
        encoder_layers_1_fc1_add_output_0 = opset20.Add(
            encoder_layers_1_fc1_bias, encoder_layers_1_fc1_matmul_output_0
        )
        encoder_layers_1_activation_fn_gelu_output_0 = opset20.Gelu(
            encoder_layers_1_fc1_add_output_0, approximate="none"
        )
        encoder_layers_1_fc2_matmul_output_0 = opset20.MatMul(
            encoder_layers_1_activation_fn_gelu_output_0, matmul_288
        )
        encoder_layers_1_fc2_add_output_0 = opset20.Add(
            encoder_layers_1_fc2_bias, encoder_layers_1_fc2_matmul_output_0
        )
        encoder_layers_1_add_1_output_0 = opset20.Add(
            encoder_layers_1_self_attn_layer_norm_layernormalization_output_0,
            encoder_layers_1_fc2_add_output_0,
        )
        encoder_output = opset20.LayerNormalization(
            encoder_layers_1_add_1_output_0,
            encoder_layers_1_final_layer_norm_weight,
            encoder_layers_1_final_layer_norm_bias,
            axis=-1,
            epsilon=9.999999747378752e-06,
        )
        return encoder_output

    return main_graph.to_model_proto()


def make_model_with_random_weights():
    encoder_embed_tokens_weight = np.random.rand(1024, 16).astype(np.float32)
    encoder_embed_positions_weight = np.random.rand(102, 16).astype(np.float32)
    encoder_layers_0_self_attn_k_proj_bias = np.random.rand(16).astype(np.float32)
    encoder_layers_0_self_attn_layer_norm_weight = np.random.rand(16).astype(np.float32)
    encoder_layers_0_fc1_bias = np.zeros((4), dtype=np.float32)

    matmul_257 = np.random.rand(16, 16).astype(np.float32)
    matmul_267 = np.random.rand(16, 16).astype(np.float32)
    matmul_268 = np.random.rand(16, 16).astype(np.float32)
    matmul_270 = np.random.rand(16, 16).astype(np.float32)
    matmul_271 = np.random.rand(16, 4).astype(np.float32)
    matmul_272 = np.random.rand(4, 16).astype(np.float32)
    matmul_273 = np.random.rand(16, 16).astype(np.float32)
    matmul_283 = np.random.rand(16, 16).astype(np.float32)
    matmul_284 = np.random.rand(16, 16).astype(np.float32)
    matmul_286 = np.random.rand(16, 16).astype(np.float32)
    matmul_287 = np.random.rand(16, 4).astype(np.float32)
    matmul_288 = np.random.rand(4, 16).astype(np.float32)

    model = make_model(
        encoder_embed_positions_weight=encoder_embed_positions_weight,
        encoder_embed_tokens_weight=encoder_embed_tokens_weight,
        encoder_layers_0_self_attn_k_proj_bias=encoder_layers_0_self_attn_k_proj_bias,
        encoder_layers_0_self_attn_layer_norm_weight=encoder_layers_0_self_attn_layer_norm_weight,
        encoder_layers_0_fc1_bias=encoder_layers_0_fc1_bias,
        matmul_257=matmul_257,
        matmul_267=matmul_267,
        matmul_268=matmul_268,
        matmul_270=matmul_270,
        matmul_271=matmul_271,
        matmul_272=matmul_272,
        matmul_273=matmul_273,
        matmul_283=matmul_283,
        matmul_284=matmul_284,
        matmul_286=matmul_286,
        matmul_287=matmul_287,
        matmul_288=matmul_288,
    )
    return model


class _BartEncoderTest:
    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            model_proto = make_model_with_random_weights()
            model = ir.serde.deserialize_model(model_proto)
            self._onnx_model = model
        return self._onnx_model

    def get_ort_inputs(self):
        if not hasattr(self, "_ort_inputs"):
            inputs = {
                "input_ids": np.random.randint(0, 1024, (1, 16)).astype(np.int64),
            }
            self._ort_inputs = inputs
        return self._ort_inputs


def bart_encoder_test():
    return _BartEncoderTest()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import unittest

import numpy as np
import onnx
import onnx_ir as ir
from onnx_ir.passes.common import onnx_checker

from onnxscript.rewriter import MatchingTracer, MatchStatus, RewriteRule, testing
from onnxscript.rewriter.rules.common import _remove_optional_bias
from onnxscript.rewriter.rules.common._remove_optional_bias import (
    remove_optional_bias_from_conv_rule,
    remove_optional_bias_from_conv_transpose_rule,
    remove_optional_bias_from_gemm_rule,
    remove_optional_bias_from_qlinear_conv_rule,
)


class _RemoveOptionalBiasTestBase(unittest.TestCase):
    @property
    def rng(self):
        return np.random.default_rng(20251016)

    def clone_model(self, model: ir.Model) -> ir.Model:
        return ir.from_proto(ir.to_proto(model))

    def _get_test_model(
        self,
        op_type: str,
        input_shape: ir.Shape,
        weight_shape: ir.Shape,
        zero_bias: bool,
        attributes=None,
    ):
        tape = ir.tape.Tape()
        bias_shape = weight_shape[1] if op_type == "ConvTranspose" else weight_shape[0]
        output_shape = ir.Shape(("?",) * input_shape.rank())

        x = ir.val("X", shape=input_shape, type=ir.TensorType(ir.DataType.FLOAT))

        w = tape.initializer(
            ir.tensor(self.rng.uniform(-0.5, 0.5, weight_shape).astype(np.float32), name="W")
        )

        if zero_bias:
            bias = np.zeros(bias_shape, dtype=np.float32)
        else:
            bias = self.rng.uniform(-0.5, 0.5, bias_shape).astype(np.float32)

        b = tape.initializer(ir.tensor(bias, name="B"))
        y = tape.op(
            op_type,
            inputs=[x, w, b],
            attributes=attributes,
            output=ir.val("Y", shape=output_shape, type=ir.TensorType(ir.DataType.FLOAT)),
        )

        # Build the model
        ir_model = ir.Model(
            ir.Graph(
                inputs=[x],
                outputs=[y],
                nodes=tape.nodes,
                initializers=tape.initializers,
                opset_imports={"": 20},
                name="test_model",
            ),
            ir_version=10,
        )
        onnx_checker.CheckerPass(True)(ir_model)
        return ir_model

    def run_test(
        self,
        base_model: ir.Model,
        input_shape: tuple,
        input_dtype=np.float32,
    ):
        updated_model = self.clone_model(base_model)
        count = _remove_optional_bias.rules.apply_to_model(updated_model)

        # Check rule is applied
        self.assertEqual(count, 1)

        # Check number of inputs is reduced
        self.assertEqual(
            len(updated_model.graph[0].inputs), len(base_model.graph[0].inputs) - 1
        )

        # Prepare inputs
        inputs = (self.rng.random(input_shape).astype(input_dtype),)

        # Check inference
        testing.assert_numerically_equal(base_model, updated_model, inputs)

        # Validate serialized model
        output_model_proto = ir.serde.serialize_model(updated_model)
        onnx.checker.check_model(output_model_proto, full_check=True)

    def run_failed_condition_test(
        self,
        base_model: ir.Model,
        rewrite_rule: RewriteRule,
        expected_message: str,
    ):
        onnx_checker.CheckerPass(True)(base_model)

        updated_model = self.clone_model(base_model)
        tracer = MatchingTracer()
        count = rewrite_rule.apply_to_model(updated_model, tracer=tracer)

        # Check that the model is unchanged
        self.assertEqual(count, 0)

        # Check that the error message is the expected one
        tracer_match = tracer.best_matches_map[rewrite_rule][0]
        self.assertEqual(tracer_match.status.value, MatchStatus.CONDITION_FAILED)
        self.assertRegex(tracer_match.match_result.reason, expected_message)


class RemoveOptionalBiasGemmTest(_RemoveOptionalBiasTestBase):
    def test_successful_remove_optional_bias_gemm(self):
        input_shape = (512, 256)
        base_model = self._get_test_model(
            op_type="Gemm",
            input_shape=ir.Shape(input_shape),
            weight_shape=ir.Shape((64, 256)),
            zero_bias=True,
            attributes={"transB": 1},
        )
        self.run_test(base_model, input_shape)

    def test_fail_remove_optional_bias_gemm(self):
        input_shape = (512, 256)
        base_model = self._get_test_model(
            op_type="Gemm",
            input_shape=ir.Shape(input_shape),
            weight_shape=ir.Shape((64, 256)),
            zero_bias=False,
            attributes={"transB": 1},
        )
        self.run_failed_condition_test(
            base_model, remove_optional_bias_from_gemm_rule, "Bias is not all zeros."
        )


class RemoveOptionalBiasGonvTest(_RemoveOptionalBiasTestBase):
    def test_successful_remove_optional_bias_conv(self):
        input_shape = (1, 3, 32, 32)
        base_model = self._get_test_model(
            op_type="Conv",
            input_shape=ir.Shape(input_shape),
            weight_shape=ir.Shape((16, 3, 3, 3)),
            zero_bias=True,
            attributes={"strides": (2, 2)},
        )
        self.run_test(base_model, input_shape)

    def test_fail_remove_optional_bias_conv(self):
        input_shape = (1, 3, 32, 32)
        base_model = self._get_test_model(
            op_type="Conv",
            input_shape=ir.Shape(input_shape),
            weight_shape=ir.Shape((16, 3, 3, 3)),
            zero_bias=False,
        )
        self.run_failed_condition_test(
            base_model, remove_optional_bias_from_conv_rule, "Bias is not all zeros."
        )


class RemoveOptionalBiasGonvTransposeTest(_RemoveOptionalBiasTestBase):
    def test_successful_remove_optional_bias_conv_transpose(self):
        input_shape = (1, 3, 32, 32)
        base_model = self._get_test_model(
            op_type="ConvTranspose",
            input_shape=ir.Shape(input_shape),
            weight_shape=ir.Shape((3, 16, 3, 3)),
            zero_bias=True,
        )
        self.run_test(base_model, input_shape)

    def test_fail_remove_optional_bias_conv_transpose(self):
        input_shape = (1, 3, 32, 32)
        base_model = self._get_test_model(
            op_type="ConvTranspose",
            input_shape=ir.Shape(input_shape),
            weight_shape=ir.Shape((3, 16, 3, 3)),
            zero_bias=False,
        )
        self.run_failed_condition_test(
            base_model, remove_optional_bias_from_conv_transpose_rule, "Bias is not all zeros."
        )


class RemoveOptionalBiasQLinearConvTest(_RemoveOptionalBiasTestBase):
    def _get_test_model(self, zero_bias):
        if zero_bias:
            bias = np.zeros((16,), dtype=np.int32)
        else:
            bias = self.rng.uniform(-5, 5, (16,)).astype(np.int32)

        w = ir.tensor(self.rng.uniform(-5, 5, (16, 3, 3, 3)).astype(np.uint8), name="W")
        b = ir.tensor(bias, name="B")

        model = ir.from_onnx_text(
            """
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (uint8[N, 3, 32, 32] X) => (uint8 [N, ?, ?, ?] Y)
            <uint8[16, 3, 3, 3] W, int32[16] B, float x_scale = {1.5}, uint8 x_zero_point = {123},
            float w_scale = {1.5}, uint8 w_zero_point = {123},
            float y_scale = {1.5}, uint8 y_zero_point = {123}>
            {
                Y = QLinearConv(X, x_scale, x_zero_point, W, w_scale, w_zero_point, y_scale, y_zero_point, B)
            }
        """,
            initializers=[w, b],
        )
        onnx_checker.CheckerPass(True)(model)
        return model

    def test_successful_remove_optional_bias_qlinear_conv(self):
        input_shape = (1, 3, 32, 32)
        base_model = self._get_test_model(zero_bias=True)
        self.run_test(base_model, input_shape, np.uint8)

    def test_fail_remove_optional_bias_qlinear_conv(self):
        base_model = self._get_test_model(zero_bias=False)
        self.run_failed_condition_test(
            base_model, remove_optional_bias_from_qlinear_conv_rule, "Bias is not all zeros."
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import Any

import numpy as np
import onnx
import onnx.reference
import onnxruntime as ort

from onnxscript import ir


def generate_random_inputs(model: onnx.ModelProto) -> dict[str, Any]:
    feeds: dict[str, Any] = {}
    for input in model.graph.input:
        input_type = input.type.tensor_type
        shape = tuple(input_type.shape.dim)
        if not all(hasattr(d, "dim_value") for d in shape):
            raise ValueError(f"Input {input.name} has dynamic shape dimensions.")
        shape = tuple(d.dim_value for d in shape)
        if input_type.elem_type == onnx.TensorProto.FLOAT:
            if shape:
                feeds[input.name] = np.random.randn(*shape).astype(np.float32)
            else:
                feeds[input.name] = np.random.randn(1).astype(np.float32)
        else:
            raise ValueError(f"Not implemented for input type {input_type.elem_type}")
    return feeds


def assert_numerically_equal(
    original_model_proto: onnx.ModelProto | ir.Model,
    rewritten_model_proto: onnx.ModelProto | ir.Model,
    args: tuple[Any, ...] | dict[str, Any] | None = None,
    ort_optimization_level: ort.GraphOptimizationLevel = ort.GraphOptimizationLevel.ORT_ENABLE_ALL,
    rtol: float = 1,
    atol: float = 1e-3,
    use_reference: bool = False,
):
    """Assert that the two models are numerically equal.

    Args:
        original_model_proto: The original model proto or ir.Model.
        rewritten_model_proto: The rewritten by the rules model proto or ir.Model.
        args: The positional arguments to pass to the model.
        ort_optimization_level: Onnxruntime optimization level.
        rtol: Relative tolerance.
        atol: Absolute tolerance.
        use_reference: If True, use ONNX reference implementation instead of ONNXRuntime.
    """

    if isinstance(original_model_proto, ir.Model):
        original_model_proto = ir.serde.serialize_model(original_model_proto)
    if isinstance(rewritten_model_proto, ir.Model):
        rewritten_model_proto = ir.serde.serialize_model(rewritten_model_proto)

    if args is None:
        original_proto_ort_inputs = generate_random_inputs(original_model_proto)
        the_rewritten_proto_ort_inputs = original_proto_ort_inputs
    elif isinstance(args, dict):
        original_proto_ort_inputs = args
        the_rewritten_proto_ort_inputs = args
    else:
        original_proto_ort_inputs = {
            k.name: v for k, v in zip(original_model_proto.graph.input, args)
        }
        the_rewritten_proto_ort_inputs = {
            k.name: v for k, v in zip(rewritten_model_proto.graph.input, args)
        }

    if use_reference:
        # Use ONNX reference implementation
        original_evaluator = _reference_session(
            original_model_proto.SerializeToString(), ort_optimization_level
        )
        original_outputs = original_evaluator.run(None, original_proto_ort_inputs)

        rewritten_evaluator = _reference_session(
            rewritten_model_proto.SerializeToString(), ort_optimization_level
        )
        the_rewritten_outputs = rewritten_evaluator.run(None, the_rewritten_proto_ort_inputs)
    else:
        # Use ONNXRuntime
        original_proto_ort_inference_session = _ort_session_initializer(
            original_model_proto.SerializeToString(), ort_optimization_level
        )
        run_options = ort.RunOptions()
        run_options.log_severity_level = 3  # 3: Error
        original_outputs = original_proto_ort_inference_session.run(
            None, original_proto_ort_inputs, run_options=run_options
        )

        the_rewritten_proto_ort_inference_session = _ort_session_initializer(
            rewritten_model_proto.SerializeToString(), ort_optimization_level
        )
        the_rewritten_outputs = the_rewritten_proto_ort_inference_session.run(
            None, the_rewritten_proto_ort_inputs, run_options=run_options
        )

    np.testing.assert_allclose(
        original_outputs, the_rewritten_outputs, rtol=rtol, atol=atol, equal_nan=True
    )


def _ort_session_initializer(
    model: str | bytes, ort_optimization_level: ort.GraphOptimizationLevel
) -> ort.InferenceSession:
    """Initialize an ONNX Runtime inference session with the specified model."""
    import onnxruntime as ort

    session_options = ort.SessionOptions()
    session_options.log_severity_level = 3  # 3: Error
    session_options.graph_optimization_level = ort_optimization_level
    possible_providers = (
        "CUDAExecutionProvider",
        "CPUExecutionProvider",
    )
    available_providers = set(ort.get_available_providers())
    providers = [
        provider for provider in possible_providers if provider in available_providers
    ]
    return ort.InferenceSession(model, providers=providers, sess_options=session_options)


def _reference_session(
    model: str | bytes, ort_optimization_level: ort.GraphOptimizationLevel
) -> onnx.reference.ReferenceEvaluator:
    """Initialize an ONNX reference evaluator with the specified model."""
    # Parse the model from bytes if needed
    if isinstance(model, (str, bytes)):
        model_proto = onnx.load_from_string(model)
    else:
        model_proto = model

    # Note: ort_optimization_level is ignored for reference implementation
    # as it doesn't have equivalent optimization levels
    return onnx.reference.ReferenceEvaluator(model_proto)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Basic types for the pattern matching and rewriter API."""

from __future__ import annotations

import dataclasses
import enum
from collections import defaultdict
from typing import TYPE_CHECKING, Any, MutableSequence, Sequence, Union

from onnxscript import ir

if TYPE_CHECKING:
    import onnxscript.rewriter._pattern_ir as _pattern_ir
    import onnxscript.rewriter._rewrite_rule as _rewrite_rule


class MatchFailureInfo:
    """Encapsulates information about a pattern match failure."""

    def __init__(
        self,
        reason: str = "",
        *failure_source: ir.Node | ir.Value,
    ):
        self.reason = reason
        self.failure_sources: tuple[ir.Node | ir.Value, ...] = failure_source
        assert all(isinstance(item, (ir.Node, ir.Value)) for item in failure_source), (
            f"All items in failure_source must be ir.Node or ir.Value, got {[type(item) for item in failure_source]}"
        )

    def __str__(self):
        return f"MatchFailureInfo(reason={self.reason!r}, failure_sources={self.failure_sources!r})"


class MatchFailureError(MatchFailureInfo, Exception):
    """Exception raised when a pattern match fails.

    This makes it easier to handle match failures in a compositional way,
    for example, during the condition-checking phase of a pattern match.
    It allows us to define utility functions without having to check for
    and propagate match failures explicitly.
    """

    def __init__(
        self,
        reason: str = "",
        *failure_source: ir.Node | ir.Value,
    ):
        MatchFailureInfo.__init__(self, reason, *failure_source)
        Exception.__init__(self, reason)


class MatchResult:
    """The state object used by the pattern-matching algorithm.

    A match can either succeed or fail.
    If it succeeds, it returns a list of nodes that matched the pattern
    and a set of bindings for the variables in the pattern.

    Example:
    ::
        def pattern(x, shape1, shape2):
            t1 = op.Reshape(x, shape1)
            t2 = op.Reshape(t1, shape2)
            return t2
    The above pattern matches a sequence of two Reshape ops.
    The matched_nodes will contain the two Reshape ops, and the bindings will
    contain the values that are bound to the variables `x`, `shape1`, and `shape2`.
    """

    def __init__(self) -> None:
        # We use a stack of partial matches to handle OR patterns that require backtracking.
        self._partial_matches: list[PartialMatchResult] = [PartialMatchResult()]

    def __repr__(self) -> str:
        """Returns a string representation of the match result."""
        if not self._partial_matches:
            return "MatchResult()"
        return (
            f"MatchResult(success={bool(self)}, reason={self.reason!r}, nodes={self.nodes!r})"
        )

    @property
    def _current_match(self) -> PartialMatchResult:
        """Returns the current match result."""
        return self._partial_matches[-1]

    def enter_new_match(self) -> None:
        """Starts a new sub-match to try out one of multiple alternatives."""
        match = PartialMatchResult()
        self._partial_matches.append(match)

    def abandon_current_match(self) -> PartialMatchResult:
        """Abandons the current alternative due to failure."""
        if len(self._partial_matches) < 2:
            raise ValueError("No match to abandon.")
        return self._partial_matches.pop()

    def merge_current_match(self) -> None:
        """Merges a successful sub-match for an alternative with the parent one."""
        if len(self._partial_matches) < 2:
            raise ValueError("No match to merge.")
        current_match = self._partial_matches.pop()
        previous_match = self._partial_matches[-1]
        if not current_match:
            raise ValueError("Current match is not successful.")
        # Merge the two matches.
        previous_match.merge(current_match)

    def __bool__(self) -> bool:
        """Returns True if the current match is successful."""
        return bool(self._current_match)

    def fail(
        self,
        reason: str = "",
        failure_source: Union[ir.Node, ir.Value, list[Union[ir.Node, ir.Value]]] | None = None,
    ) -> MatchResult:
        self._current_match.fail(reason, failure_source)
        return self

    @property
    def reason(self) -> str:
        """Returns the reason for the failure."""
        return self._current_match.reason

    @property
    def nodes(self) -> Sequence[ir.Node]:
        """Returns the list of nodes that matched the pattern."""
        return self._current_match.nodes

    def bind_node(self, pattern_node: _pattern_ir.NodePattern, node: ir.Node):
        """Binds a pattern node to a matched node."""
        self.add_node(node)
        self._current_match.node_bindings[pattern_node] = node

    def add_node(self, node: ir.Node) -> None:
        """Adds a node to the list of matched nodes."""
        self._current_match.add_node(node)

    def bind_value(self, pattern_value: _pattern_ir.ValuePattern, value: Any) -> bool:
        var_name = pattern_value.name
        # TODO(rama): Simplify the following. We currently bind values to
        # pattern variables in two different ways: via their name, or via the
        # pattern-value itself.
        if var_name is None:
            for match in self._partial_matches:
                if pattern_value in match.value_bindings:
                    # TODO(rama): Use appropriate equality-check here.
                    if match.value_bindings[pattern_value] == value:
                        return True
                    self._current_match.fail(
                        f"Binding failure: {pattern_value} bound to two different values.",
                        [match.value_bindings[pattern_value], value],
                    )
                    return False
            self._current_match.value_bindings[pattern_value] = value
            return True
        return self.bind(var_name, value)

    def bind(self, var: str, value: Any) -> bool:
        for match in self._partial_matches:
            if var in match.bindings:
                # TODO(rama): Use appropriate equality-check here.
                if match.bindings[var] == value:
                    return True
                self._current_match.fail(
                    f"Binding failure: {var} bound to two different values.",
                    [match.bindings[var], value],
                )
                return False
        self._current_match.bindings[var] = value
        return True

    @property
    def bindings(self) -> dict[str, Any]:
        """Returns the bindings for the pattern variables."""
        if len(self._partial_matches) > 1:
            raise ValueError("Bindings can be accessed only at the top-level match.")
        return self._current_match.bindings

    @property
    def value_bindings(self) -> dict[_pattern_ir.ValuePattern, ir.Value]:
        """Returns the bindings for the value variables."""
        if len(self._partial_matches) > 1:
            raise ValueError("Value bindings can be accessed only at the top-level match.")
        return self._current_match.value_bindings

    @property
    def node_bindings(self) -> dict[_pattern_ir.NodePattern, ir.Node]:
        """Returns the bindings for the node variables."""
        if len(self._partial_matches) > 1:
            raise ValueError("Node bindings can be accessed only at the top-level match.")
        return self._current_match.node_bindings

    @property
    def outputs(self) -> MutableSequence[ir.Value]:
        """Returns the list of output values that matched the pattern."""
        if len(self._partial_matches) > 1:
            raise ValueError("Outputs can be accessed only at the top-level match.")
        return self._current_match.outputs

    @property
    def failure_nodes_and_values(self) -> list[Union[ir.Node, ir.Value]]:
        """Returns the nodes and values that caused the failure."""
        return self._current_match._failure_nodes_and_values

    def lookup_node(self, pattern_node: _pattern_ir.NodePattern) -> ir.Node | None:
        """Looks up the node that matched the given pattern node."""
        for match in self._partial_matches:
            if pattern_node in match.node_bindings:
                return match.node_bindings[pattern_node]
        return None

    def num_matched_nodes(self) -> int:
        """Returns the number of nodes matched so far."""
        return sum(len(match.node_bindings) for match in self._partial_matches)


class PartialMatchResult:
    """The state object used by the pattern-matching algorithm for a sub-match."""

    def __init__(self) -> None:
        self._success: bool = True
        # For a successful match, _matched_nodes is a list of values that matched the pattern.
        # These include the internal nodes of the pattern that were matched, but not
        # the leaves (sub-trees) that match against the variables in the pattern.
        # These represent the values that will be replaced by the replacement pattern.
        self._matched_nodes: MutableSequence[ir.Node] = []
        # For a successful match, bindings is a dictionary of mapping pattern-variable-names
        # to values.
        self._bindings: dict[str, Any] = {}
        self._value_bindings: dict[_pattern_ir.ValuePattern, ir.Value] = {}
        self._node_bindings: dict[_pattern_ir.NodePattern, ir.Node] = {}

        self._outputs: list[ir.Value] = []
        # For a failed match, _reason is a string that describes the reason for the failure.
        self._reason: str = ""
        # Track the node(s) or value(s) that caused the failure.
        self._failure_nodes_and_values: list[Union[ir.Node, ir.Value]] = []

    def __bool__(self):
        return self._success

    def fail(
        self,
        reason: str = "",
        failure_source: Union[ir.Node, ir.Value, list[Union[ir.Node, ir.Value]]] | None = None,
    ) -> None:
        self._success = False
        self._reason = reason
        if failure_source is not None:
            if isinstance(failure_source, list):
                self._failure_nodes_and_values.extend(failure_source)
            else:
                self._failure_nodes_and_values.append(failure_source)

    @property
    def reason(self) -> str:
        return self._reason

    @property
    def nodes(self) -> Sequence[ir.Node]:
        return tuple(self._matched_nodes)

    def add_node(self, node: ir.Node) -> None:
        """Adds a node to the list of matched nodes."""
        self._matched_nodes.append(node)

    @property
    def bindings(self) -> dict[str, Any]:
        return self._bindings

    @property
    def value_bindings(self) -> dict[_pattern_ir.ValuePattern, ir.Value]:
        return self._value_bindings

    @property
    def outputs(self) -> MutableSequence[ir.Value]:
        return self._outputs

    @property
    def node_bindings(self) -> dict[_pattern_ir.NodePattern, ir.Node]:
        return self._node_bindings

    def merge(self, other: PartialMatchResult) -> None:
        """Merges a successful sub-match for an alternative with the parent one."""
        if self._success and other._success:
            # Merge the two successful matches. Matching algorithm responsible for ensuring
            # that the two matches are compatible. No need to check for conflicts here.
            self._bindings.update(other._bindings)
            self._matched_nodes.extend(other.nodes)
            # Note: outputs should be set only at end of the (top-level) match. There
            # should be no outputs in the sub-match.
            assert not other._outputs
        else:
            # This should not happen currently.
            raise NotImplementedError("Merging failed matches is not yet supported.")


class MatchStatus(enum.IntEnum):
    """The status of a pattern-matching operation."""

    NO_MATCH = 0  # No successful match found for entire pattern graph
    CONDITION_FAILED = 1  # Subsequent validation check failed
    REPLACEMENT_FAILED = 2  # Replacement subgraph could not be created
    SUCCESS = 3  # A successful match was found


@dataclasses.dataclass
class MatchInfo:
    """The status of a pattern-matching operation. An extension of MatchResult."""

    match_result: MatchResult
    root_node: ir.Node
    container: ir.Graph | ir.Function
    status: MatchStatus

    def score(self) -> int:
        """Return a score for the match."""
        return len(self.match_result.nodes) + int(self.status.value) * 100

    def print(self):
        separator = "-" * 80
        print(separator)
        print(f"Status: {self.status.name}")
        if self.status != MatchStatus.SUCCESS:
            reason = self.match_result.reason
            if reason:
                if self.status == MatchStatus.CONDITION_FAILED:
                    print(f"Graph matching failed due to failing check condition : {reason}")
                else:
                    print(f"Graph matching failed: {reason}")
            else:
                print("Graph matching failed.")
            failure_nodes_and_values = self.match_result.failure_nodes_and_values
            print("Failure at or around nodes/values:")
            if failure_nodes_and_values:
                for failure_cause in failure_nodes_and_values:
                    failure_cause.display()
        print("Matched nodes:")
        import onnxscript.rewriter._ir_utils as ir_utils

        ir_utils.display_nodes(self.match_result.nodes)
        print(separator)


class MatchContext:
    """A read-only context containing information about a pattern match.

    This class captures information about the context describing a match to a given pattern,
    providing access to the model, graph/function, root node, output values, and all
    nodes of the matching subgraph.
    """

    def __init__(
        self,
        model: ir.Model,
        graph_or_function: ir.Graph | ir.Function,
        root: ir.Node,
        match_result: MatchResult,
    ) -> None:
        """Initialize the pattern match context.

        Args:
            model: The model being matched.
            graph_or_function: The graph or function being matched.
            root: The root node of the matching subgraph.
            match_result: The match result containing matched nodes and outputs.
        """
        self._model = model
        self._graph_or_function = graph_or_function
        self._root = root
        self._match_result = match_result

    @property
    def model(self) -> ir.Model:
        """The model being matched."""
        return self._model

    @property
    def graph_or_function(self) -> ir.Graph | ir.Function:
        """The graph or function being matched."""
        return self._graph_or_function

    @property
    def root(self) -> ir.Node:
        """The root node of the matching subgraph."""
        return self._root

    @property
    def output_values(self) -> Sequence[ir.Value]:
        """The output values of the matching subgraph."""
        return self._match_result.outputs

    @property
    def nodes(self) -> Sequence[ir.Node]:
        """All the nodes of the matching subgraph."""
        return self._match_result.nodes

    def display(self, *, in_graph_order: bool = True) -> None:
        """Display the nodes in the pattern match context.

        Args:
            in_graph_order: If True, display nodes in the order they appear in the
                graph/function. If False, display nodes in the order they appear
                in the match result.
        """
        nodes = self.nodes
        if not nodes:
            return

        if in_graph_order:
            # Display nodes in same order as in graph/function
            for node in self._graph_or_function:
                if node in nodes:
                    node.display()
        else:
            # Display nodes in match order
            for node in nodes:
                node.display()


class MatchingTracer:
    """A debugging helper class to trace the matching of a pattern against a graph.

    This is used to track the best matches found for each rule, and to report the
    results at the end of the matching.
    """

    def __init__(self) -> None:
        self._best_matches_map: dict[_rewrite_rule.RewriteRule, list[MatchInfo]] = defaultdict(
            list
        )

    @property
    def best_matches_map(self) -> dict[_rewrite_rule.RewriteRule, list[MatchInfo]]:
        return self._best_matches_map

    def log(
        self,
        rule: _rewrite_rule.RewriteRule,
        container: ir.Graph | ir.Function,
        node: ir.Node,
        match_result: MatchResult,
        status: MatchStatus,
    ) -> None:
        this_match = MatchInfo(match_result, node, container, status)
        this_score = this_match.score()
        if this_score == 0:
            return
        best_matches = self._best_matches_map[rule]
        if best_matches:
            if this_score < best_matches[0].score():
                return
            if this_score > best_matches[0].score():
                best_matches.clear()
        best_matches.append(this_match)

    def report(self) -> None:
        best_score = 0
        for rule, matches in self._best_matches_map.items():
            if not matches:
                continue
            if matches[0].score() > best_score:
                best_score = matches[0].score()
                best_match = matches[0]
                best_rule = rule

        if best_score > 0:
            print(f"Rule: {best_rule}")
            best_match.print()
        else:
            print("No matches found.")



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from __future__ import annotations

from typing import Any

from onnxscript._internal import builder as _builder
from onnxscript.nn import _module_list


class Sequential(_module_list.ModuleList):
    """A sequential container that calls children in order, mirroring ``torch.nn.Sequential``.

    Children are registered with string keys ``"0"``, ``"1"``, etc., just like
    ``ModuleList``. The ``forward`` method passes the output of each child as
    the input to the next.

    Example::

        class SiLU(Module):
            def forward(self, op, x):
                return op.Mul(x, op.Sigmoid(x))

        # Produces parameter names: "mod.0.weight", "mod.0.bias"
        # SiLU at index 0 has no parameters.
        mod = Sequential(SiLU(), Linear(4, 4))

        # Calling mod(op, x) is equivalent to:
        #   x = silu(op, x)
        #   x = linear(op, x)
    """

    def __init__(self, *modules: _module_list.Module) -> None:
        super().__init__(modules)

    def _set_name(self, name: str) -> None:
        """Set this container's name. Children keep simple ``"0"``, ``"1"`` names.

        Unlike ``ModuleList._set_name`` which fully qualifies children (used
        when ModuleList is iterated externally), Sequential is called via
        ``__call__`` which already pushes its own name onto the builder stack.
        Children must keep simple keys to avoid double-prefixing.
        """
        object.__setattr__(self, "_name", name)
        for key, child in self._modules.items():
            child._set_name(key)  # pylint: disable=protected-access

    def _register_child(self, key: str, module: _module_list.Module) -> None:
        """Register a child module under the given string key.

        Unlike ``ModuleList._register_child`` which qualifies the child name
        with the parent name, Sequential keeps children with simple index
        names because ``__call__`` already pushes the Sequential's own name.
        """
        if module._name is None:  # pylint: disable=protected-access
            object.__setattr__(module, "_name", key)
        self._modules[key] = module
        object.__setattr__(self, key, module)

    def forward(self, op: _builder.OpBuilder, input: Any) -> Any:  # pylint: disable=redefined-builtin
        """Run each child module sequentially, passing output to the next.

        Mirrors ``torch.nn.Sequential.forward``: each child receives exactly
        one positional argument (the output of the previous child).
        """
        if len(self) == 0:
            raise RuntimeError("Cannot call forward on an empty Sequential container")
        for module in self:
            input = module(op, input)
        return input

    def __repr__(self) -> str:
        lines = ["Sequential("]
        for name, module in self._modules.items():
            mod_repr = repr(module).replace("\n", "\n  ")
            lines.append(f"  ({name}): {mod_repr}")
        lines.append(")")
        return "\n".join(lines)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import contextlib
import io
import logging
import unittest

import numpy as np
import onnx.checker
import onnx.parser

import onnxscript.optimizer
import onnxscript.rewriter
from onnxscript import FLOAT, ir, script
from onnxscript import opset17 as op
from onnxscript.rewriter import pattern
from onnxscript.rewriter.rules.common import _cast_constant_of_shape

logger = logging.getLogger(__name__)


class ReciprocalMulTest(unittest.TestCase):
    def rule(self) -> pattern.RewriteRule:
        def reciprocal_mul_pattern(op, x, y):
            return (1 / x) * y

        def div(op, x, y):
            return op.Div(y, x)

        return pattern.RewriteRule(reciprocal_mul_pattern, div)

    def test_single_match(self):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[N] z)
            {
                c1 = Constant<value_float = 1.0>()
                t1 = Div(c1, x)
                z1 = Mul(t1, y)
                z = Identity(z1)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = self.rule().apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 3)

    def test_failed_match(self):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[N] z)
            {
                c1 = Constant<value_float = 0.9>()
                t1 = Div(c1, x)
                z1 = Mul(t1, y)
                z = Identity(z1)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = self.rule().apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 4)

        # Test verbose output produces something:
        # TODO(rama): Need a better way to test this.
        # Well-defined error-codes and messages would be helpful.

        buffer = io.StringIO()
        with contextlib.redirect_stdout(buffer):
            self.rule().apply_to_model(model, verbose=5)
        out = buffer.getvalue()
        self.assertIn("Match failed", out)

    def test_multiple_matches(self):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[N] z)
            {
                # {c1, t1, z1} is a valid match
                # {c2, t2, z2} is a valid match
                # {c3, t3, z3} is a match, but cannot be replaced since t3 has other-uses.
                c1 = Constant<value_float = 1.0>()
                c2 = Constant<value_float = 1.0>()
                t2 = Div(c2, y)
                t1 = Div(c1, x)
                z1 = Mul(t1, y)
                z2 = Mul(t2, z1)

                c3 = Constant<value_float = 1.0>()
                t3 = Div(c3, x)
                z3 = Mul(t3, y)
                reuse_t3 = Div(t3, x)
                z = Add(z2, reuse_t3)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = self.rule().apply_to_model(model)
        self.assertEqual(count, 2)
        self.assertEqual(len(model.graph), 9)


class FastGeluTest(unittest.TestCase):
    def rule(self) -> pattern.RewriteRule:
        def fast_gelu_pattern1(op, x):
            b = 0.044715
            c = 0.79788
            tanh = op.Tanh(c * (x + (x**3) * b))
            return (1.0 + tanh) * (0.5 * x)

        def fast_gelu(op, x):
            return op.FastGelu(x, _domain="com.microsoft")

        return pattern.RewriteRule(fast_gelu_pattern1, fast_gelu)

    def long_form_rule(self) -> pattern.RewriteRule:
        def fast_gelu_pattern1_long(op, x):
            three = pattern.Constant(3)
            x_cube = op.Pow(x, three)
            b = pattern.Constant(0.044715)
            x_cube_mul_b = op.Mul(x_cube, b)  # support OR op.Mul(B, x_cube)
            sum_ = op.Add(x, x_cube_mul_b)
            c = pattern.Constant(0.79788)
            c_times_sum = op.Mul(c, sum_)
            tanh = op.Tanh(c_times_sum)
            one = pattern.Constant(1.0)
            one_plus_tanh = op.Add(one, tanh)
            half = pattern.Constant(0.5)
            half_x = op.Mul(half, x)
            return op.Mul(one_plus_tanh, half_x)

        def fast_gelu(op, x):
            return op.FastGelu(x, _domain="com.microsoft")

        return pattern.RewriteRule(fast_gelu_pattern1_long, fast_gelu)

    def _check(self, rule):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[N] z)
            {
                three = Constant <value_int=3>()
                x_cube = Pow(x, three)
                B = Constant <value_float=0.044715>()
                x_cube_mul_B = Mul(x_cube, B)
                sum = Add(x, x_cube_mul_B)
                C = Constant <value_float=0.79788>()
                C_times_sum = Mul(C, sum)
                tanh = Tanh(C_times_sum)
                one = Constant <value_float=1.0> ()
                one_plus_tanh = Add(one, tanh)
                half = Constant <value_float=0.5> ()
                half_x = Mul(half, x)
                z = Mul(one_plus_tanh, half_x)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = rule.apply_to_model(model)
        self.assertEqual(count, 1)
        # 5 Constant nodes and 1 FastGelu node
        self.assertEqual(len(model.graph), 6)

    def test_short_rule(self):
        self._check(self.rule())

    def test_long_rule(self):
        self._check(self.long_form_rule())


class ConcatTest(unittest.TestCase):
    def rule(self) -> pattern.RewriteRule:
        def concat_pattern(op, x, y, axis):
            seq = op.SequenceConstruct(x, y)
            return op.ConcatFromSequence(seq, axis=axis)

        def concat(op, x, y, axis):
            return op.Concat(x, y, axis=axis)

        return pattern.RewriteRule(concat_pattern, concat)

    def test_concat(self):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[M] z)
            {
                t = SequenceConstruct (x, y)
                z = ConcatFromSequence <axis=0> (t)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = self.rule().apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 1)

    def test_concat_in_function(self):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17, "pkg.custom": 1]>
            agraph (float[N] x, float[M] y) => (float[Z] z)
            {
                z = afunction (x, y)
            }
            <domain: "pkg.custom", opset_import: [ "" : 17]>
            afunction (x, y) => (z)
            {
                t = SequenceConstruct (x, y)
                z = ConcatFromSequence <axis=0> (t)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = self.rule().apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.functions), 1)
        self.assertEqual(len(model.functions[("pkg.custom", "afunction", "")]), 1)
        self.assertEqual(model.functions[("pkg.custom", "afunction", "")][0].op_type, "Concat")


class RewriteRuleTest(unittest.TestCase):
    def test_commute(self):
        def add_0(op, x):
            return x + 0

        def identity(op, x):
            return op.Identity(x)

        add_0_rule = pattern.RewriteRule(add_0, identity)

        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[M] z)
            {
                zero = Constant <value_float=0.0> ()
                z = Add (zero, x)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = pattern.RewriteRuleSet([add_0_rule], commute=True).apply_to_model(model)
        optimized_model = ir.serde.serialize_model(model)
        self.assertEqual(count, 1)
        nodes = optimized_model.graph.node
        self.assertEqual(len(nodes), 2)
        self.assertEqual(nodes[1].op_type, "Identity")

    def test_const_value(self):
        def reshape(op, x, newshape):
            return op.Reshape(x, newshape)

        def identity(op, x, newshape):
            del newshape  # Unused
            return op.Identity(x)

        def check_for_redundant_reshape(context, x, newshape):
            oldshape = x.shape
            newshape_const_value = newshape.const_value
            if newshape_const_value is None:
                return False

            newshape = newshape_const_value.numpy()
            newshape = newshape.tolist()

            if len(oldshape) != len(newshape):
                return False
            return all(not (d1 != d2 and d2 != -1) for d1, d2 in zip(oldshape, newshape))  # pylint: disable=consider-using-in

        rule = pattern.RewriteRule(reshape, identity, check_for_redundant_reshape)

        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[10, 20, 30] x) => (float[10, 20, 30] z)
            {
                shape = Constant <value_ints=[10, 20, 30]> ()
                z = Reshape (x, shape)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = pattern.RewriteRuleSet([rule]).apply_to_model(model)
        optimized_model = ir.serde.serialize_model(model)
        self.assertEqual(count, 1)
        nodes = optimized_model.graph.node
        self.assertEqual(len(nodes), 2)
        self.assertEqual(nodes[1].op_type, "Identity")

    def test_delayed_run_provides_correct_bindings_for_multiple_matches(self):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (int64[2] input_x) => (float16[1, 4] output, float[1, 4] output2)
            {
                constant = ConstantOfShape <value: tensor = float[1] {1.}>(input_x)
                output = Cast <to = 10> (constant)
                constant2 = ConstantOfShape <value: tensor = float[1] {1.}>(input_x)
                output2 = Cast <to = 1> (constant2)
            }
            """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _cast_constant_of_shape.rules.apply_to_model(model)
        self.assertEqual(count, 2)
        self.assertEqual(len(model.graph), 2)
        self.assertEqual(model.graph[0].attributes["value"].value.dtype, 10)
        self.assertEqual(model.graph[1].attributes["value"].value.dtype, 1)

    def test_opset_import(self):
        def add_same(op, x):
            return x + x

        def double(op, x):
            return op.Double(x, _domain="custom.domain", _version=10)

        rule = pattern.RewriteRule(add_same, double)

        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[M] z)
            {
                y = Add (x, x)
                z = Relu (y)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = pattern.RewriteRuleSet([rule], commute=True).apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(model.graph.opset_imports["custom.domain"], 10)

    def test_opset_import_in_function(self):
        def add_same(op, x):
            return x + x

        def double(op, x):
            return op.Double(x, _domain="custom.domain", _version=10)

        rule = pattern.RewriteRule(add_same, double)

        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17, "pkg.custom": 1]>
            agraph (float[N] x) => (float[M] z)
            {
                z = pkg.custom.afunction (x)
            }
            <domain: "pkg.custom", opset_import: [ "" : 17]>
            afunction (x) => (z)
            {
                y = Add (x, x)
                z = Relu (y)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = pattern.RewriteRuleSet([rule], commute=True).apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.functions), 1)
        self.assertEqual(model.graph.opset_imports["custom.domain"], 10)
        self.assertEqual(
            model.functions[("pkg.custom", "afunction", "")].opset_imports["custom.domain"], 10
        )
        onnx.checker.check_model(ir.serde.serialize_model(model))

    def test_optional_attribute(self):
        """Test rules with optional attributes."""

        def concat_pattern(op, x, y):
            seq = op.SequenceConstruct(x, y)
            result = op.ConcatFromSequence(seq, _outputs=["result"])
            return result

        def concat(op, x, y, result: ir.Value):
            node = result.producer()
            assert node is not None
            axis = node.attributes.get("axis", None)
            return op.Concat(x, y, axis=axis)

        rule = pattern.RewriteRule(concat_pattern, concat)

        # Case 1: a model with attribute axis present
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[M] z)
            {
                t = SequenceConstruct (x, y)
                z = ConcatFromSequence <axis=0> (t)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = rule.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 1)
        self.assertEqual(model.graph[0].op_type, "Concat")
        self.assertEqual(model.graph[0].attributes["axis"].value, 0)

        # Case 2: a model with attribute axis absent
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[M] z)
            {
                t = SequenceConstruct (x, y)
                z = ConcatFromSequence (t)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = rule.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 1)
        self.assertEqual(model.graph[0].op_type, "Concat")
        self.assertNotIn("axis", model.graph[0].attributes)

    def test_match_none_input(self):
        def none_pattern(op, x):
            # match against a call to Original where the first input is None
            return op.Original(None, x)

        def replacement(op, x):
            return op.Replaced(x)

        rule = pattern.RewriteRule(none_pattern, replacement)

        @script()
        def test_model(x: FLOAT[1024]) -> FLOAT[1024]:
            # Pattern should match following call
            t1 = op.Original(None, x)
            # Pattern should not match following call
            z = op.Original(t1, x)
            return z

        model_proto = test_model.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)

        count = rule.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 2)
        self.assertEqual(model.graph.node(0).op_type, "Replaced")
        self.assertEqual(model.graph.node(1).op_type, "Original")

    def test_match_optional_input(self):
        def none_pattern(op, x):
            # match against a call to Original where the first input may or may not be None
            optional_input = pattern.Var("optional_input", can_match_none=True)
            return op.Original(optional_input, x)

        def replacement(op, optional_input, x):
            if optional_input is None:
                return op.ReplacedNone(x)
            return op.ReplacedNotNone(x)

        rule = pattern.RewriteRule(none_pattern, replacement)

        @script()
        def test_model(x: FLOAT[1024]) -> FLOAT[1024]:
            # Pattern should match following call
            t1 = op.Original(None, x)
            # as well as this one
            z = op.Original(t1, x)
            return z

        model_proto = test_model.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)

        count = rule.apply_to_model(model)
        self.assertEqual(count, 2)
        self.assertEqual(len(model.graph), 2)
        self.assertEqual(model.graph.node(0).op_type, "ReplacedNone")
        self.assertEqual(model.graph.node(1).op_type, "ReplacedNotNone")

    def test_mismatched_number_of_inputs(self):
        def var_length_pattern(op):
            # match against a call to Original where the first input may or may not be None
            input1 = pattern.Var("input1", can_match_none=False)
            input2 = pattern.Var("input2", can_match_none=True)
            return op.Original(input1, input2)

        def replacement(op, input1, input2):
            return op.Replaced(input1, input2)

        rule = pattern.RewriteRule(var_length_pattern, replacement)

        @script()
        def test_model(x: FLOAT[1024], y: FLOAT[1024], z: FLOAT[1024]) -> FLOAT[1024]:
            # Pattern should NOT match following 2 calls, since pattern requires first input to be non-None
            t0 = op.Original()
            t1 = op.Original(None, x)

            # Pattern should match following 3 calls, since second input can be None
            t2 = op.Original(x)
            t3 = op.Original(x, None)
            t4 = op.Original(x, y)

            # Pattern should NOT match following call, since it has more than 2 inputs
            t5 = op.Original(x, y, z)
            return op.All(t0, t1, t2, t3, t4, t5)

        model_proto = test_model.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)

        count = rule.apply_to_model(model)
        self.assertEqual(count, 3)
        self.assertEqual(len(model.graph), 7)
        self.assertEqual(
            [n.op_type for n in model.graph],
            ["Original", "Original", "Replaced", "Replaced", "Replaced", "Original", "All"],
        )

    def test_graph_visitor(self):
        class ReplaceFoo(pattern.RewriteRuleClassBase):
            def __init__(self):
                super().__init__()
                self.replacement = None

            def pattern(self, op):
                return op.Foo()

            def rewrite(self, op):
                if self.replacement is None:
                    self.replacement = op.Bar()
                return self.replacement

        rule = ReplaceFoo.rule()

        @script()
        def test_model(x: FLOAT[1024]) -> FLOAT[1024]:
            # Pattern should match following call
            t1 = op.Foo()
            # as well as this one
            t2 = op.Foo()
            z = op.Add(t1, t2)
            return z

        model_proto = test_model.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)

        count = rule.apply_to_model(model)
        self.assertEqual(count, 2)
        self.assertEqual(len(model.graph), 2)
        self.assertEqual(model.graph.node(0).op_type, "Bar")
        self.assertEqual(model.graph.node(1).op_type, "Add")

    def test_debug_mode(self):
        def source_pattern(op, x):
            t1 = op.Abs(x)
            t2 = op.Neg(t1)
            t3 = op.Exp(t2)
            return t3

        def replacement(op, x):
            return op.Something(x)

        rule = pattern.RewriteRule(source_pattern, replacement)

        @script()
        def test_model(x: FLOAT[1024]) -> FLOAT[1024]:
            a2 = op.Abs(x)  # match-1 fails here
            a3 = op.Exp(a2)  # match-1 starts here
            b1 = op.Neg(a3)  # match-2 fails here
            b2 = op.Neg(b1)  # match-2 (partially) succeeds here
            b3 = op.Exp(b2)  # match-2 starts here
            return b3

        model_proto = test_model.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)

        tracer = pattern.MatchingTracer()
        count = rule.apply_to_model(model, tracer=tracer)
        self.assertEqual(count, 0)
        best_matches = tracer.best_matches_map[rule]
        self.assertEqual(len(best_matches), 1)
        best_match = best_matches[0]
        self.assertEqual(best_match.status.value, pattern.MatchStatus.NO_MATCH)
        self.assertIn("OpType mismatch: expected Abs, got Neg", best_match.match_result.reason)

    def test_new_initializer(self):
        def source_pattern(op, x, y):
            return op.Gemm(x, op.Transpose(y))

        def check(context, x, y):
            return y.const_value is not None

        def replacement(op, x, y):
            tensor = y.const_value
            name = y.name + "_transposed"
            transposed = ir.tensor(tensor.numpy().T, name=name)
            initializer = op.initializer(transposed)
            return op.Gemm(x, initializer)

        rule = pattern.RewriteRule(source_pattern, replacement, check)

        y_value = np.random.rand(8, 4).astype(np.float32)

        @script()
        def test_model(x: FLOAT[16, 8]) -> FLOAT[16, 4]:
            y = op.Constant(value=y_value)
            return op.Gemm(x, op.Transpose(y))

        model_proto = test_model.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        rule.apply_to_model(model)
        self.assertEqual(len(model.graph.initializers), 1)
        last_node = model.graph[-1]
        self.assertEqual(len(last_node.inputs), 2)
        init_name = last_node.inputs[1].name
        self.assertIn(init_name, model.graph.initializers)
        self.assertIs(last_node.inputs[1], model.graph.initializers[init_name])

    def test_extract_function(self):
        def source_pattern(op, x, y, z):
            sum = op.Add(x, y)
            return op.Mul(sum, z)

        def replacement(op, x, y, z):
            return op.AddMul(x, y, z, _domain="some.domain")

        rule = pattern.RewriteRule(source_pattern, replacement, as_function=True)

        @script()
        def test_model(x: FLOAT[1024], y: FLOAT[1024], z: FLOAT[1024]) -> FLOAT[1024]:
            return op.Mul(op.Add(x, y), z)

        model_proto = test_model.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        rule.apply_to_model(model)
        self.assertEqual(len(model.functions), 1)
        self.assertEqual(len(model.graph), 1)
        call_node = model.graph.node(0)
        self.assertEqual(call_node.domain, "some.domain")
        self.assertEqual(call_node.op_type, "AddMul")
        function_id = call_node.op_identifier()
        self.assertIn(function_id, model.functions)
        function = model.functions[function_id]
        self.assertEqual([x.op_type for x in function], ["Add", "Mul"])
        onnxscript.optimizer.inline(model)
        self.assertEqual([x.op_type for x in model.graph], ["Add", "Mul"])

    def test_extract_function_with_attr(self):
        def source_pattern(op, x, y):
            sum = op.Add(x, y)
            return op.Transpose(sum, perm=[1, 0])

        def replacement(op, x, y):
            return op.AddTranspose(x, y, _domain="some.domain")

        rule = pattern.RewriteRule(source_pattern, replacement, as_function=True)

        @script()
        def test_model(x: FLOAT[1024, 512], y: FLOAT[1024, 512]) -> FLOAT[512, 1024]:
            return op.Transpose(op.Add(x, y), perm=[1, 0])

        model_proto = test_model.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        rule.apply_to_model(model)
        self.assertEqual(len(model.functions), 1)
        self.assertEqual(len(model.graph), 1)
        call_node = model.graph.node(0)
        self.assertEqual(call_node.domain, "some.domain")
        self.assertEqual(call_node.op_type, "AddTranspose")
        function_id = call_node.op_identifier()
        self.assertIn(function_id, model.functions)
        function = model.functions[function_id]
        self.assertEqual([x.op_type for x in function], ["Add", "Transpose"])
        transpose_node = function[1]
        self.assertEqual(list(transpose_node.attributes["perm"].value), [1, 0])
        onnxscript.optimizer.inline(model)
        self.assertEqual([x.op_type for x in model.graph], ["Add", "Transpose"])

    def test_extract_repeated_function(self):
        def source_pattern(op, x, y, z):
            sum = op.Add(x, y)
            return op.Mul(sum, z)

        def replacement(op, x, y, z):
            return op.AddMul(x, y, z, _domain="some.domain")

        rule = pattern.RewriteRule(source_pattern, replacement, as_function=True)

        @script()
        def test_model(x: FLOAT[1024], y: FLOAT[1024], z: FLOAT[1024]) -> FLOAT[1024]:
            t1 = op.Mul(op.Add(x, y), z)
            t2 = op.Mul(op.Add(t1, y), z)
            return t2

        model_proto = test_model.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        rule.apply_to_model(model)
        self.assertEqual(len(model.functions), 2)
        self.assertEqual(len(model.graph), 2)
        for call_node in model.graph:
            self.assertEqual(call_node.domain, "some.domain")
            self.assertEqual(call_node.op_type, "AddMul")
            function_id = call_node.op_identifier()
            self.assertIn(function_id, model.functions)
        onnxscript.optimizer.inline(model)
        self.assertEqual([x.op_type for x in model.graph], ["Add", "Mul", "Add", "Mul"])

    def test_any_value(self):
        def source_pattern(op, x):
            return op.Add(x, op.Mul(0, pattern.ANY_VALUE))

        def replacement(op, x):
            return op.Identity(x)

        rule = pattern.RewriteRule(source_pattern, replacement)

        @script()
        def test_model(x: FLOAT[1024], y: FLOAT[1024]) -> FLOAT[1024]:
            zero = op.Constant(value_float=0.0)
            return op.Add(x, op.Mul(zero, y))

        model_proto = test_model.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        self.assertEqual([x.op_type for x in model.graph], ["Constant", "Mul", "Add"])
        rule.apply_to_model(model)
        self.assertEqual(len(model.graph), 2)
        self.assertEqual([x.op_type for x in model.graph], ["Constant", "Identity"])

    def test_or_pattern(self):
        def source_pattern(op, x, y, bias):
            t1 = op.MatMul(x, y)
            t2 = op.Add(t1, bias)
            t1_or_t2 = pattern.OrValue([t1, t2], tag_var="has_bias", tag_values=[False, True])
            return op.Relu(t1_or_t2)

        def replacement(op, x, y, bias, has_bias):
            if has_bias:
                return op.WithBias(x, y, bias)
            else:
                return op.WithoutBias(x, y)

        rule = pattern.RewriteRule(source_pattern, replacement)

        @script()
        def test_model1(x: FLOAT[16, 32], y: FLOAT[32, 16]) -> FLOAT[16, 16]:
            return op.Relu(op.MatMul(x, y))

        model_proto = test_model1.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        rule.apply_to_model(model)
        self.assertEqual([x.op_type for x in model.graph], ["WithoutBias"])

        @script()
        def test_model2(x: FLOAT[16, 32], y: FLOAT[32, 16], bias: FLOAT[16]) -> FLOAT[16, 16]:
            return op.Relu(op.Add(op.MatMul(x, y), bias))

        model_proto = test_model2.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        rule.apply_to_model(model)
        self.assertEqual([x.op_type for x in model.graph], ["WithBias"])

    def test_backtracking_pattern(self):
        def source_pattern(op, x, y, bias):
            t1 = op.MatMul(x, y)
            choice1 = op.Add(t1, bias)
            choice2 = op.Add(bias, t1)
            t2 = pattern.OrValue([choice1, choice2])
            return op.Relu(t2)

        def replacement(op, x, y, bias):
            return op.GemmRelu(x, y, bias)

        rule = pattern.RewriteRule(source_pattern, replacement)

        @script()
        def test_model1(x: FLOAT[16, 32], y: FLOAT[32, 16], bias: FLOAT[16]) -> FLOAT[16, 16]:
            return op.Relu(op.Add(op.MatMul(x, y), bias))

        model_proto = test_model1.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        rule.apply_to_model(model)
        self.assertEqual([x.op_type for x in model.graph], ["GemmRelu"])
        self.assertEqual([x.name for x in model.graph.node(0).inputs], ["x", "y", "bias"])

        @script()
        def test_model2(x: FLOAT[16, 32], y: FLOAT[32, 16], bias: FLOAT[16]) -> FLOAT[16, 16]:
            return op.Relu(op.Add(bias, op.MatMul(x, y)))

        model_proto = test_model2.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        rule.apply_to_model(model)
        self.assertEqual([x.op_type for x in model.graph], ["GemmRelu"])
        self.assertEqual([x.name for x in model.graph.node(0).inputs], ["x", "y", "bias"])

    def test_or_pattern_return_value(self):
        """Test that an OrValue can be used as a return value from the source pattern."""

        def source_pattern(op, x, y):
            choice1 = op.Add(x, y)
            choice2 = op.Mul(x, y)
            t = pattern.OrValue([choice1, choice2])
            z = op.Relu(t)
            return z, t

        def replacement(op, x, y):
            z, t = op.ReluPlus(x, y, _outputs=2)
            return z, t

        rule = pattern.RewriteRule(source_pattern, replacement)

        @script()
        def test_model1(x: FLOAT[16, 32], y: FLOAT[16, 32]) -> FLOAT[16, 32]:
            return op.Relu(op.Add(x, y))

        model_proto = test_model1.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        rule.apply_to_model(model)
        self.assertEqual([x.op_type for x in model.graph], ["ReluPlus"])


class ValueNodeCheckersTest(unittest.TestCase):
    """Test value/node level checkers functionality."""

    def test_pattern_match_with_node_checker(self):
        """Test Pattern.match with node-level checker."""

        def shape_node_checker(context, node):
            return node.attributes.get_int("start", 0) == 0

        # Create a pattern that matches Shape operations with a node checker
        def shape_pattern(op, x):
            return op.Shape(x, _check=shape_node_checker)

        # Create the pattern
        rule_pattern = pattern.Pattern(shape_pattern)

        # Create a model with multiple Shape nodes with different start attributes
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N, M] x) => (int64[2] z1, int64[2] z2, int64[1] z3)
            {
                z1 = Shape(x)
                z2 = Shape <start: int = 0>(x)
                z3 = Shape <start: int = 1>(x)
            }
            """
        )
        model = ir.serde.deserialize_model(model_proto)

        # Find the Shape nodes in the model
        nodes = list(model.graph)
        shape_node_no_attr = nodes[0]  # Shape without start attribute
        shape_node_start_0 = nodes[1]  # Shape with start=0
        shape_node_start_1 = nodes[2]  # Shape with start=1

        self.assertEqual(shape_node_no_attr.op_type, "Shape")
        self.assertEqual(shape_node_start_0.op_type, "Shape")
        self.assertEqual(shape_node_start_1.op_type, "Shape")

        # Test case 1: Shape without start attribute (should match, default is 0)
        match_result = rule_pattern.match(model, model.graph, shape_node_no_attr)
        self.assertTrue(bool(match_result))

        # Test case 2: Shape with start=0 (should match)
        match_result = rule_pattern.match(model, model.graph, shape_node_start_0)
        self.assertTrue(bool(match_result))

        # Test case 3: Shape with start=1 (should not match)
        match_result = rule_pattern.match(model, model.graph, shape_node_start_1)
        self.assertFalse(bool(match_result))

    def test_pattern_match_with_value_checker(self):
        """Test Pattern.match with value-level checker."""

        def is_positive_constant(context, value: ir.Value):
            if value.const_value is not None:
                # Get the numpy array from const_value
                numpy_array = value.const_value.numpy()

                # Check if it represents a single value and is positive
                if numpy_array.size != 1:
                    return False

                return float(numpy_array.item()) > 0

            return False

        # Create a pattern with value checker using callable directly
        def add_pattern(op, x, y):
            # Use callable as input to create ValuePattern with checker
            return op.Add(is_positive_constant, y)

        # Create the pattern
        rule_pattern = pattern.Pattern(add_pattern)

        # Create a model with several calls to Add:
        # - one with first parameter non-constant
        # - one with first parameter a positive constant
        # - one with first parameter a negative constant
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[N] z1, float[N] z2, float[N] z3)
            {
                pos_const = Constant <value_float = 2.5> ()
                neg_const = Constant <value_float = -1.5> ()
                z1 = Add(x, y)           # non-constant first parameter
                z2 = Add(pos_const, y)   # positive constant first parameter
                z3 = Add(neg_const, y)   # negative constant first parameter
            }
            """
        )
        model = ir.serde.deserialize_model(model_proto)

        # Apply constant propagation to set const_value fields
        onnxscript.optimizer.basic_constant_propagation(model.graph.all_nodes())

        # Find the Add nodes in the model
        add_nodes = [node for node in model.graph if node.op_type == "Add"]
        self.assertEqual(len(add_nodes), 3)

        # Test case 1: Non-constant first parameter - should not match
        match_result = rule_pattern.match(model, model.graph, add_nodes[0])
        self.assertFalse(bool(match_result))

        # Test case 2: Positive constant first parameter - should match
        match_result = rule_pattern.match(model, model.graph, add_nodes[1])
        self.assertTrue(bool(match_result))
        self.assertEqual(len(match_result.nodes), 1)
        self.assertGreaterEqual(len(match_result.value_bindings), 1)

        # Test case 3: Negative constant first parameter - should not match
        match_result = rule_pattern.match(model, model.graph, add_nodes[2])
        self.assertFalse(bool(match_result))

    def test_rule_name_metadata(self):
        """Test that RewriteRule carries name metadata."""

        class ReciprocalMulRule(pattern.RewriteRuleClassBase):
            def __init__(self, name: str | None = None):
                super().__init__(name)

            def pattern(self, op, x, y):
                return (1 / x) * y

            def rewrite(self, op, x, y):
                return op.Div(y, x)

        @script()
        def test_script(x: FLOAT[1024], y: FLOAT[1024]) -> FLOAT[1024]:
            return op.Mul(op.Div(op.Constant(value_float=1.0), x), y)

        rule = ReciprocalMulRule.rule(name="ReciprocalMulToDiv")
        model_proto = test_script.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        count = rule.apply_to_model(model)
        self.assertEqual(count, 1)
        for node in model.graph:
            if node.op_type == "Div":
                tag = onnxscript.rewriter.RULE_NAME_TAG
                self.assertEqual(node.metadata_props.get(tag), "ReciprocalMulToDiv")

        # By default, the rule name is the class name (if not provided)
        rule = ReciprocalMulRule.rule()
        model_proto = test_script.to_model_proto()
        model = ir.serde.deserialize_model(model_proto)
        count = rule.apply_to_model(model)
        self.assertEqual(count, 1)
        for node in model.graph:
            if node.op_type == "Div":
                tag = onnxscript.rewriter.RULE_NAME_TAG
                self.assertEqual(node.metadata_props.get(tag), "ReciprocalMulRule")


class PatternBuilderTest(unittest.TestCase):
    def test_pattern_builder_context(self):
        builder = pattern.OpsetPatternBuilder("", True)
        with pattern.pattern_builder(builder):
            x = builder.Op1()
            y = builder.Op2(x)
            z = x + y
            w = builder.Op3(z)
            _ = z * w
        ops = [x.op_type for x in builder.nodes()]
        self.assertEqual(ops, ["Op1", "Op2", "Add", "Op3", "Mul"])


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import onnx_ir as ir
import onnx_ir.passes.common as common_passes

import onnxscript.rewriter.ort_fusions.fused_matmul_rule_sets as fused_matmul_rule_sets
import onnxscript.rewriter.ort_fusions.shape_optimization as shape_optimization
from onnxscript.optimizer import optimize
from onnxscript.rewriter import rewrite
from onnxscript.rewriter.ort_fusions import (
    instance_to_group_normalization,
    softmax,
)
from onnxscript.rewriter.ort_fusions.attention import fuse_attention
from onnxscript.rewriter.ort_fusions.bias_gelu import fuse_bias_gelu
from onnxscript.rewriter.ort_fusions.cos_sin_cache import fuse_cos_sin_cache
from onnxscript.rewriter.ort_fusions.erfgelu import fuse_erfgelu
from onnxscript.rewriter.ort_fusions.gelu import fuse_gelu
from onnxscript.rewriter.ort_fusions.gqa import fuse_gqa
from onnxscript.rewriter.ort_fusions.gqa_packed_qkv import fuse_qkv_gqa
from onnxscript.rewriter.ort_fusions.mha import fuse_mha1, fuse_mha2
from onnxscript.rewriter.ort_fusions.mha_bias import fuse_mha_bias
from onnxscript.rewriter.ort_fusions.mha_scale import fuse_mha_scale
from onnxscript.rewriter.ort_fusions.rms_normalization import fuse_rms_normalization
from onnxscript.rewriter.ort_fusions.rotary_embedding import (
    fuse_partial_rotary_embedding,
    fuse_rotary_embedding,
)
from onnxscript.rewriter.ort_fusions.sdpa import fuse_sdpa
from onnxscript.rewriter.ort_fusions.sdpa_via_mha import replace_sdpa_by_mha
from onnxscript.rewriter.ort_fusions.skip_normalization import (
    fuse_skip_layer_normalization,
    fuse_skip_rms_normalization,
)
from onnxscript.rewriter.rules.common import _gemm_to_matmul_add

ORT_PATTERN_REWRITE_RULES = [
    *softmax.rules.rules,
    *instance_to_group_normalization.rules.rules,
    # NOTE: group normalization merge silu should be applied after instance to group normalization
    # *group_normalization_merge_silu.rules.rules,
    *fused_matmul_rule_sets.fused_matmul_rule_sets(),
]


# Preliminary optimizations before applying the transformer fusions.
# TODO: There are some potential redundancies below. Can be targeted for optimization
# once we have robust fusion.
def _pre_optimize(model: ir.Model) -> ir.Model:
    # TODO: Do we need this dependence on ONNX's partial-data-propagation? There are some
    # extra shape-propagation and partial-data-propagation rules in ONNX that are not yet
    # incorporated in our optimizer.
    common_passes.ShapeInferencePass()(model)
    optimize(model)
    shape_optimization.rules.apply_to_model(model)
    optimize(model)
    return model


def fuse_xformers(model: ir.Model, debug: bool = False) -> tuple[ir.Model, dict[str, int]]:
    """
    Apply transformer-specific fusions to the given model.

    Args:
        model: The input ONNX model represented as an `ir.Model`.
        debug: If debug is True, enable pattern matching tracer for debugging.

    Returns:
        A tuple containing:
        - The optimized `ir.Model` after applying transformer-specific fusions.
        - A dictionary with a count of each of the fusions applied.
    """
    fusion_count = dict()

    model = _pre_optimize(model)

    def fuse(func, **kwargs):
        return func(model, debug=debug, **kwargs)

    fusion_count["erf_gelu"] = fuse(fuse_erfgelu)
    fusion_count["rms_normalization"] = fuse(fuse_rms_normalization)
    fusion_count["skip_layer_normalization"] = fuse(fuse_skip_layer_normalization)
    fusion_count["skip_rms_normalization"] = fuse(fuse_skip_rms_normalization)
    fusion_count["rotary_embedding"] = fuse(fuse_rotary_embedding)
    fusion_count["cos_sin_cache"] = fuse(fuse_cos_sin_cache)
    common_passes.CommonSubexpressionEliminationPass()(model)
    fusion_count["partial_rotary_embedding"] = fuse(fuse_partial_rotary_embedding)

    # We apply shape inference after the SDPA fusion as new nodes are added
    # in the rewrite rule for certain patterns of SDPA.
    fusion_count["sdpa"] = fuse(fuse_sdpa, apply_shape_inference=True)

    fusion_count["gqa"] = fuse(fuse_gqa)
    fusion_count["packed_qkv_for_gqa"] = fuse(fuse_qkv_gqa)
    fusion_count["mha1"] = fuse(fuse_mha1)
    fusion_count["mha2"] = fuse(fuse_mha2)
    fusion_count["mha_scale"] = fuse(fuse_mha_scale)
    if (fusion_count["mha1"] == 0) and (fusion_count["mha2"] == 0):
        fusion_count["mha_bias"] = 0
        fusion_count["attention"] = 0
    else:
        fusion_count["mha_bias"] = fuse(fuse_mha_bias)
        fusion_count["attention"] = fuse(fuse_attention)
    fusion_count["gelu"] = fuse(fuse_gelu)
    fusion_count["bias_gelu"] = fuse(fuse_bias_gelu)
    fusion_count["sdpa_via_mha"] = fuse(replace_sdpa_by_mha)
    # Finally: inline any intermediate fusion functions introduced that were not
    # consumed by other fusions, and eliminate any remaining unused nodes.
    optimize(model)
    return model, fusion_count


def optimize_for_ort(
    model: ir.Model,
    config_name: str | None = None,
    *,
    debug: bool = False,
    clear_metadata: bool = False,
) -> tuple[ir.Model, dict[str, int]]:
    """
    Optimize the model for ORT backend.

    TODO: config_name is not used yet. It should be used to select the appropriate
    optimization configuration (for an EP). Currently, a default implementation is used.

    Args:
        model: The model to optimize.
        config_name: The name of the configuration to use for optimization.
            Typically it identifies the Execution Provider (EP) to optimize for.
            If None, the default configuration will be used.
        debug: If debug is True, enable pattern matching tracer for debugging.
        clear_metadata: If True, clear metadata and doc strings from the model.

    Returns:
        A tuple containing:
        - The optimized `ir.Model` after applying transformer-specific fusions.
        - A dictionary with a count of each of the fusions applied.
    """
    rewrite(model, [_gemm_to_matmul_add.gemm_to_matmul_add_rule])
    model, fusion_count = fuse_xformers(
        model,
        debug=debug,
    )
    # Apply the ORT pattern rewrite rules.
    rewrite(model, ORT_PATTERN_REWRITE_RULES)

    passes = ir.passes.Sequential(
        # Apply the ORT optimization passes.
        # https://github.com/microsoft/onnxruntime/blob/74dcf7e296639095dfa55d31336998b6f719ed76/onnxruntime/python/tools/transformers/dynamo_onnx_helper.py#L172
        # https://github.com/microsoft/onnxruntime/blob/74dcf7e296639095dfa55d31336998b6f719ed76/onnxruntime/python/tools/transformers/dynamo_onnx_helper.py#L139
        common_passes.LiftConstantsToInitializersPass(lift_all_constants=False, size_limit=1),
        common_passes.RemoveInitializersFromInputsPass(),
        common_passes.ShapeInferencePass(),
    )
    assert passes.in_place
    result = passes(model)
    assert result.model is model

    if clear_metadata:
        common_passes.ClearMetadataAndDocStringPass()(model)

    return model, fusion_count



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern

# Fusions for RotaryEmbedding:
# Fuse computation patterns seen in HF transformer models for RotaryEmbedding
# and map them to ONNX opset 23 RotaryEmbedding op.

# Basic pattern: For example, see
# https://github.com/huggingface/transformers/blob/541bed22d6e4f97946a3a7d74f7e1a353e58643b/src/transformers/models/llama/modeling_llama.py#L104
#    def rotate_half(x):
#        """Rotates half the hidden dims of the input."""
#        x1 = x[..., : x.shape[-1] // 2]
#        x2 = x[..., x.shape[-1] // 2 :]
#        return torch.cat((-x2, x1), dim=-1)
# and
#        q_embed = (q * cos) + (rotate_half(q) * sin)


def _rotate_half_pattern(op, x, start1, end1, start2, end2):
    # Slice(input, starts, ends, axes, steps)
    x1 = op.Slice(x, start1, end1, [3], [1])
    x2 = op.Slice(x, start2, end2, [3], [1])
    minus_x2 = op.Neg(x2)
    rotated_x = op.Concat(minus_x2, x1, axis=-1)
    return rotated_x


class RotaryEmbedding23Fusion(pattern.RewriteRuleClassBase):
    def __init__(self):
        super().__init__(name="RotaryEmbedding23", remove_nodes=False)

    def pattern(self, op, x, freqs, start1, end1, start2, end2, one1, one2):
        freqs_repeated = op.Concat(freqs, freqs, axis=-1)
        cos = op.Cos(freqs_repeated)
        sin = op.Sin(freqs_repeated)
        cos_4d = op.Unsqueeze(cos, one1)
        sin_4d = op.Unsqueeze(sin, one2)
        return x * cos_4d + _rotate_half_pattern(op, x, start1, end1, start2, end2) * sin_4d

    def check(self, op, x, start1, end1, start2, end2, one1, one2, **_) -> pattern.MatchResult:  # type: ignore[name-defined]
        check_result = pattern.MatchResult()

        if not _ir_utils.is_singleton_value(one1, 1):
            return check_result.fail("Unsqueeze axes is not [1]", one1)
        if not _ir_utils.is_singleton_value(one2, 1):
            return check_result.fail("Unsqueeze axes is not [1]", one2)

        # x needs to be a 4D tensor with known last dimension size (== head_size) and known second dimension (num_heads)
        if x is None or x.shape is None or len(x.shape) != 4:
            return check_result.fail("Input is not known to be a 4D tensor.", x)
        if not isinstance(x.shape[1], int):
            return check_result.fail("Input dimension 1 (num_heads) is not static.", x)
        head_size = x.shape[3]
        if not isinstance(head_size, int):
            return check_result.fail("Head size is not static.", x)
        half_head_size = head_size // 2

        # Check that x is being split into two equal halves of size half_head_size
        if not (
            _ir_utils.is_singleton_value(start1, 0)
            and _ir_utils.is_singleton_value(end1, half_head_size)
            and _ir_utils.is_singleton_value(start2, half_head_size)
            and _ir_utils.is_singleton_value(end2, lambda x: x >= head_size)
        ):
            return check_result.fail(
                "x is not being split into two equal halves of size half_head_size."
            )
        return check_result

    def rewrite(self, op, x, freqs, **_):
        num_heads = x.shape[1]
        cos = op.Cos(freqs)
        sin = op.Sin(freqs)
        return op.RotaryEmbedding(
            x,
            cos,
            sin,
            interleaved=0,
            num_heads=num_heads,
        )


# Extensions for partial rotary embedding fusion: with partial rotary embedding,
# embedding is applied only to the first part of the input, and the second part is left unchanged,
# as captured in the pattern below.

MAX_INT64 = 9223372036854775807


class PartialRotaryEmbedding23Fusion(pattern.RewriteRuleClassBase):
    def pattern(self, op, x, end1, start2):
        x_part_1 = op.Slice(x, [0], end1, [3], [1])
        x_part_2 = op.Slice(x, start2, [MAX_INT64], [3], [1])
        x_part_1_rope = op.RotaryEmbedding(
            x_part_1,
            _allow_other_inputs=True,
            _allow_other_attributes=True,
            _outputs=["x_part_1_rope"],
        )
        return op.Concat(x_part_1_rope, x_part_2, axis=-1)

    def check(self, op, x, end1, start2, x_part_1_rope, **_) -> pattern.MatchResult:  # type: ignore[name-defined]
        check_result = pattern.MatchResult()
        end1_value = _ir_utils.get_singleton_value(end1)
        start2_value = _ir_utils.get_singleton_value(start2)
        if not isinstance(end1_value, int) or not isinstance(start2_value, int):
            return check_result.fail("Unable to validate slice start/end values.")
        if end1_value != start2_value:
            return check_result.fail(
                "The end1 value of first slice and start2 value of second slice are not equal."
            )
        rotary_embedding_attributes = x_part_1_rope.producer().attributes
        if "rotary_embedding_dim" in rotary_embedding_attributes:
            return check_result.fail("rotary_embedding_dim attribute already specified.")
        if (
            "interleaved" in rotary_embedding_attributes
            and rotary_embedding_attributes["interleaved"].value != 0
        ):
            return check_result.fail("interleaved is not equal to 0.")
        return check_result

    def rewrite(self, op, x, end1, x_part_1_rope, **_):
        # Create a modified version of the RotaryEmbedding op:
        rotary_embedding_dim = _ir_utils.get_singleton_value(end1)
        original_node = x_part_1_rope.producer()
        inputs = list(original_node.inputs)
        inputs[0] = x
        attrs = dict(original_node.attributes)
        attrs["rotary_embedding_dim"] = rotary_embedding_dim
        return op.RotaryEmbedding(
            *inputs,
            **attrs,
        )


_rule = RotaryEmbedding23Fusion.rule()

_partial_embedding_rule = PartialRotaryEmbedding23Fusion.rule()

rotary_embedding_rules = pattern.RewriteRuleSet([_rule])

partial_embedding_rules = pattern.RewriteRuleSet([_partial_embedding_rule])

fuse_rotary_embedding = _fusion_utils.apply_fusion_rules(rotary_embedding_rules)

fuse_partial_rotary_embedding = _fusion_utils.apply_fusion_rules(partial_embedding_rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Tests for the remove-Expand-before-binary-op fusion rule."""

from __future__ import annotations

import unittest

import numpy as np
import onnx
import onnx.helper
import onnx.numpy_helper
import onnx.reference
import onnx.shape_inference
import parameterized

import onnxscript.ir as ir
from onnxscript.rewriter.rules.common import _remove_expand_before_binary_op as mod


def _run_model(model: ir.Model, feeds: dict) -> list:
    """Run a model using the ONNX reference evaluator."""
    proto = ir.to_proto(model)
    ref = onnx.reference.ReferenceEvaluator(proto)
    return ref.run(None, feeds)


class RemoveExpandBeforeBinaryOpTest(unittest.TestCase):
    """Tests for _remove_expand_before_binary_op rules."""

    def _apply_and_check(
        self,
        model_text: str,
        expected_count: int,
        expected_op_types: list[str],
    ) -> ir.Model:
        """Helper: apply the rules and verify the result."""
        model = ir.from_onnx_text(model_text)
        count = mod.expand_before_binary_op_rules.apply_to_model(model)
        self.assertEqual(count, expected_count)
        actual_op_types = [node.op_type for node in model.graph]
        self.assertEqual(actual_op_types, expected_op_types)
        return model

    # ------------------------------------------------------------------
    # Cases where the Expand should be removed
    # ------------------------------------------------------------------

    @parameterized.parameterized.expand(
        [
            ("Add",),
            ("Sub",),
            ("Mul",),
            ("Div",),
        ]
    )
    def test_expand_first_input_same_shape_is_removed(self, op_type: str):
        """Expand producing same shape as input should be removed from BinaryOp."""
        model_text = f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[3, 4] x, float[3, 4] y) => (float[3, 4] output)
            <int64[2] shape = {{3, 4}}>
            {{
                expanded = Expand(x, shape)
                output = {op_type}(expanded, y)
            }}
        """
        model = self._apply_and_check(model_text, 1, [op_type])

        # Verify numerical correctness
        x = np.random.randn(3, 4).astype(np.float32)
        y = np.random.randn(3, 4).astype(np.float32)
        original = ir.from_onnx_text(model_text)
        expected = _run_model(original, {"x": x, "y": y})
        got = _run_model(model, {"x": x, "y": y})
        np.testing.assert_allclose(got[0], expected[0], rtol=1e-5)

    def test_expand_first_input_broadcast_covered_by_other_input(self):
        """Expand from [3, 4] to [4, 3, 4] can be removed when y has shape [4, 3, 4]."""
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[3, 4] x, float[4, 3, 4] y) => (float[4, 3, 4] output)
            <int64[3] shape = {4, 3, 4}>
            {
                expanded = Expand(x, shape)
                output = Add(expanded, y)
            }
        """
        model = self._apply_and_check(model_text, 1, ["Add"])

        x = np.random.randn(3, 4).astype(np.float32)
        y = np.random.randn(4, 3, 4).astype(np.float32)
        original = ir.from_onnx_text(model_text)
        expected = _run_model(original, {"x": x, "y": y})
        got = _run_model(model, {"x": x, "y": y})
        np.testing.assert_allclose(got[0], expected[0], rtol=1e-5)

    def test_expand_second_input_is_removed(self):
        """Expand on the second input of a binary op should be removed."""
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[4, 3, 4] x, float[3, 4] y) => (float[4, 3, 4] output)
            <int64[3] shape = {4, 3, 4}>
            {
                expanded = Expand(y, shape)
                output = Mul(x, expanded)
            }
        """
        model = self._apply_and_check(model_text, 1, ["Mul"])

        x = np.random.randn(4, 3, 4).astype(np.float32)
        y = np.random.randn(3, 4).astype(np.float32)
        original = ir.from_onnx_text(model_text)
        expected = _run_model(original, {"x": x, "y": y})
        got = _run_model(model, {"x": x, "y": y})
        np.testing.assert_allclose(got[0], expected[0], rtol=1e-5)

    def test_expand_with_broadcast_compatible_other_input(self):
        """Expand from [3] to [4, 3] can be removed when y has shape [4, 1]."""
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[3] x, float[4, 1] y) => (float[4, 3] output)
            <int64[2] shape = {4, 3}>
            {
                expanded = Expand(x, shape)
                output = Add(expanded, y)
            }
        """
        model = self._apply_and_check(model_text, 1, ["Add"])

        x = np.random.randn(3).astype(np.float32)
        y = np.random.randn(4, 1).astype(np.float32)
        original = ir.from_onnx_text(model_text)
        expected = _run_model(original, {"x": x, "y": y})
        got = _run_model(model, {"x": x, "y": y})
        np.testing.assert_allclose(got[0], expected[0], rtol=1e-5)

    def test_expand_sub_first_input_is_removed(self):
        """Expand on the first input of Sub should be removed."""
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[3, 4] x, float[3, 4] y) => (float[3, 4] output)
            <int64[2] shape = {3, 4}>
            {
                expanded = Expand(x, shape)
                output = Sub(expanded, y)
            }
        """
        model = self._apply_and_check(model_text, 1, ["Sub"])

        x = np.random.randn(3, 4).astype(np.float32)
        y = np.random.randn(3, 4).astype(np.float32)
        original = ir.from_onnx_text(model_text)
        expected = _run_model(original, {"x": x, "y": y})
        got = _run_model(model, {"x": x, "y": y})
        np.testing.assert_allclose(got[0], expected[0], rtol=1e-5)

    def test_expand_div_second_input_is_removed(self):
        """Expand on the second input of Div should be removed."""
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[4, 3, 4] x, float[3, 4] y) => (float[4, 3, 4] output)
            <int64[3] shape = {4, 3, 4}>
            {
                expanded = Expand(y, shape)
                output = Div(x, expanded)
            }
        """
        model = self._apply_and_check(model_text, 1, ["Div"])

        x = np.random.randn(4, 3, 4).astype(np.float32)
        y = np.random.randn(3, 4).astype(np.float32) + 2.0  # avoid division by zero
        original = ir.from_onnx_text(model_text)
        expected = _run_model(original, {"x": x, "y": y})
        got = _run_model(model, {"x": x, "y": y})
        np.testing.assert_allclose(got[0], expected[0], rtol=1e-5)

    # ------------------------------------------------------------------
    # Cases where the Expand should NOT be removed
    # ------------------------------------------------------------------

    def test_expand_changes_output_shape_not_removed(self):
        """Expand that changes the output shape compared to direct broadcast must be kept."""
        # x has shape [3], expand to [4, 3], other is a scalar.
        # With expand: broadcast([4, 3], []) = [4, 3]
        # Without expand: broadcast([3], []) = [3]  <- different!
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[3] x) => (float[4, 3] output)
            <int64[2] shape = {4, 3}, float[1] one = {1.0}>
            {
                expanded = Expand(x, shape)
                output = Add(expanded, one)
            }
        """
        model = ir.from_onnx_text(model_text)
        count = mod.expand_before_binary_op_rules.apply_to_model(model)
        self.assertEqual(count, 0)

    def test_expand_target_shape_not_constant_removed_via_output_shape(self):
        """Expand with a dynamic shape is removed when the binary op output shape
        confirms the expansion is redundant.

        x=[3, 4] has no dimension equal to 1, so Expand can only output [3, 4].
        With y=[3, 4] the binary op output shape is also [3, 4], and
        broadcast([3, 4], [3, 4]) = [3, 4] matches, so the expand is provably a
        no-op and is safely removed.
        """
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[3, 4] x, float[3, 4] y, int64[2] shape) => (float[3, 4] output)
            {
                expanded = Expand(x, shape)
                output = Add(expanded, y)
            }
        """
        model = ir.from_onnx_text(model_text)
        count = mod.expand_before_binary_op_rules.apply_to_model(model)
        self.assertEqual(count, 1)
        """Expand with a symbolic x dim can be removed when y statically covers the expansion.

        x=[N], expand_shape=[3, 4], y=[3, 4]: since y provides all expand dimensions
        as known integers, the expand is redundant regardless of N's runtime value.
        """
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[3, 4] y) => (float[3, 4] output)
            <int64[2] shape = {3, 4}>
            {
                expanded = Expand(x, shape)
                output = Add(expanded, y)
            }
        """
        model = ir.from_onnx_text(model_text)
        count = mod.expand_before_binary_op_rules.apply_to_model(model)
        self.assertEqual(count, 1)

    def test_expand_with_symbolic_y_dim_not_removed(self):
        """Expand cannot be removed when y has a symbolic dim in a position where the
        expand is doing work and that symbolic dim cannot be verified to equal expand_d.
        """
        # x=[3], expand_shape=[4, 3], y=[M, 3].
        # At dim 0 (expand adds dim 4): x_d=1 (virtual), y_d=M (symbolic) -> can't verify.
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[3] x, float[M, 3] y) => (float[4, 3] output)
            <int64[2] shape = {4, 3}>
            {
                expanded = Expand(x, shape)
                output = Add(expanded, y)
            }
        """
        model = ir.from_onnx_text(model_text)
        count = mod.expand_before_binary_op_rules.apply_to_model(model)
        self.assertEqual(count, 0)

    def test_full_optimization(self):
        oh = onnx.helper
        model_proto = oh.make_model(
            oh.make_graph(
                [
                    oh.make_node("Shape", ["x"], ["n"], start=0, end=1),
                    oh.make_node("Shape", ["x"], ["b"], start=1, end=2),
                    oh.make_node("Concat", ["n", "b"], ["shape"], axis=0),
                    oh.make_node("Expand", ["x", "shape"], ["expanded"]),
                    oh.make_node("Add", ["expanded", "y1"], ["z1"]),
                    oh.make_node("Add", ["expanded", "y2"], ["z2"]),
                    oh.make_node("Add", ["expanded", "y3"], ["z3"]),
                    oh.make_node("Add", ["z1", "z2"], ["z12"]),
                    oh.make_node("Add", ["z12", "z3"], ["z"]),
                ],
                "test",
                [
                    oh.make_tensor_value_info("x", onnx.TensorProto.FLOAT, ["N", 1]),
                    oh.make_tensor_value_info("y1", onnx.TensorProto.FLOAT, [1, "B"]),
                    oh.make_tensor_value_info("y2", onnx.TensorProto.FLOAT, [1, "B"]),
                    oh.make_tensor_value_info("y3", onnx.TensorProto.FLOAT, [1, "B"]),
                ],
                [
                    oh.make_tensor_value_info("z", onnx.TensorProto.FLOAT, ["N", "B"]),
                ],
            ),
            ir_version=10,
            opset_imports=[oh.make_opsetid("", 20)],
        )
        onnx.checker.check_model(model_proto)
        # Shape inference is required so that the Expand output carries its
        # shape annotation ([N, 1]).  Without it the rule cannot verify that
        # the expansion is redundant.
        inferred_proto = onnx.shape_inference.infer_shapes(model_proto, data_prop=True)
        model = ir.serde.deserialize_model(inferred_proto)
        count = mod.expand_before_binary_op_rules.apply_to_model(model)
        self.assertEqual(count, 3)
        self.assertEqual(len(model.graph), 5)

    def test_full_optimization_more_complex(self):
        oh = onnx.helper
        onh = onnx.numpy_helper

        model_proto = oh.make_model(
            oh.make_graph(
                [
                    oh.make_node("Shape", ["x"], ["n"], start=0, end=1),
                    oh.make_node("Shape", ["x"], ["b"], start=1, end=2),
                    oh.make_node("Concat", ["n", "b"], ["shape"], axis=0),
                    oh.make_node("Add", ["shape", "one"], ["shape1"]),
                    oh.make_node("Sub", ["shape1", "one"], ["shape2"]),
                    oh.make_node("Expand", ["x", "shape2"], ["expanded"]),
                    oh.make_node("Add", ["expanded", "y1"], ["z1"]),
                    oh.make_node("Add", ["expanded", "y2"], ["z2"]),
                    oh.make_node("Add", ["expanded", "y3"], ["z3"]),
                    oh.make_node("Add", ["z1", "z2"], ["z12"]),
                    oh.make_node("Add", ["z12", "z3"], ["z"]),
                ],
                "test",
                [
                    oh.make_tensor_value_info("x", onnx.TensorProto.FLOAT, ["N", 1]),
                    oh.make_tensor_value_info("y1", onnx.TensorProto.FLOAT, [1, "B"]),
                    oh.make_tensor_value_info("y2", onnx.TensorProto.FLOAT, [1, "B"]),
                    oh.make_tensor_value_info("y3", onnx.TensorProto.FLOAT, [1, "B"]),
                ],
                [
                    oh.make_tensor_value_info("z", onnx.TensorProto.FLOAT, ["N", "B"]),
                ],
                [onh.from_array(np.array([1], dtype=np.int64), "one")],
                # Explicit shape annotations on intermediate values (as produced by
                # shape inference or by the model creator).  These allow the rule to
                # verify that the Expand is redundant without tracing the exact
                # computation that produced the shape tensor.
                value_info=[
                    oh.make_tensor_value_info("expanded", onnx.TensorProto.FLOAT, ["N", 1]),
                    oh.make_tensor_value_info("z1", onnx.TensorProto.FLOAT, ["N", "B"]),
                    oh.make_tensor_value_info("z2", onnx.TensorProto.FLOAT, ["N", "B"]),
                    oh.make_tensor_value_info("z3", onnx.TensorProto.FLOAT, ["N", "B"]),
                ],
            ),
            ir_version=10,
            opset_imports=[oh.make_opsetid("", 20)],
        )
        onnx.checker.check_model(model_proto)
        model = ir.serde.deserialize_model(model_proto)
        count = mod.expand_before_binary_op_rules.apply_to_model(model)
        self.assertEqual(count, 3)
        self.assertEqual(len(model.graph), 5)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import math
from typing import Union

import onnx_ir as ir

from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern
from onnxscript.rewriter._basics import MatchFailureError

Dim = Union[int, ir.SymbolicDim]

# This file contains a fusion rule that recognizes various patterns of scaled dot-product attention
# (SDPA) implementations and replaces them with a single SDPA op. The SDPA op is a temporary fusion
# op defined in the ai.onnxruntime._fusion domain. Subsequent fusion rules will map it into one
# of the various ops defined in ORT: MHA, GQA, or Attention depending on the input patterns.
# The SDPA is a standard scalar dot-product attention with an optional mask input and scaling factor.
# Currently, it is restricted to query, key, and values of rank 4 with shapes:
#   Query: [batch_size, num_heads, seq_len, head_size_qk]
#   Key:   [batch_size, num_heads, seq_len_kv, head_size_qk]
#          or [batch_size, seq_len_kv, num_heads, head_size_qk])
#   Value: [batch_size, num_heads, seq_len_kv, head_size_v]
# The key_format attribute indicates which of the two formats the key uses and can be either "BHSd" or "BSHd".


class SDPA(pattern.RewriteRuleClassBase):
    _scale: float | None

    def pattern(
        self,
        op,
        query,
        key,
        value,
        mask,
        query_scale,
        key_scale,
        qk_scale,
    ):
        # The last two axes of key must be transposed before computing the dot product with query.
        # Three patterns are observed in practice:

        # Pattern 1: Transpose 4D key directly: BHSd => BHdS
        key_transposed_1 = op.Transpose(key, perm=[0, 1, 3, 2])

        # Pattern 2: Transpose key after converting to 3D and then convert back to 4D: BHSd => 3D => BHdS
        key_3d = op.Reshape(key, pattern.ANY_VALUE)
        key_3d_transposed = op.Transpose(key_3d, perm=[0, 2, 1])
        key_transposed_2 = op.Reshape(key_3d_transposed, pattern.ANY_VALUE)

        # Pattern 3: This transpose is sometimes composed with an earlier transpose to convert
        # the key from BSHd format to BHSd format.
        key_transposed_3 = op.Transpose(key, perm=[0, 2, 3, 1])

        key_transposed = pattern.OrValue(
            [key_transposed_1, key_transposed_2, key_transposed_3],
            tag_var="key_format",
            tag_values=["BHSd", "BHSd", "BSHd"],
        )

        # Some implementations scale the query and key before computing the dot product
        query = pattern.OrValue(
            [
                op.Mul(query, query_scale),
                op.Div(query, query_scale),
                query,
            ],
            tag_var="query_scaling",
            tag_values=["Mul", "Div", "None"],
        )
        key_transposed = pattern.OrValue(
            [
                op.Mul(key_transposed, key_scale),
                op.Div(key_transposed, key_scale),
                key_transposed,
            ],
            tag_var="key_scaling",
            tag_values=["Mul", "Div", "None"],
        )

        attn_score = op.MatMul(query, key_transposed)

        # Some implementations scale the dot product.
        attn_score = pattern.OrValue(
            [
                op.Mul(attn_score, qk_scale),
                op.Div(attn_score, qk_scale),
                attn_score,
            ],
            tag_var="qk_scaling",
            tag_values=["Mul", "Div", "None"],
        )

        # Some implementations add a mask to the dot product.
        masked_attn_score = op.Add(attn_score, mask)
        attn_score = pattern.OrValue(
            [masked_attn_score, attn_score], tag_var="has_mask", tag_values=[True, False]
        )

        attn_weight = op.Softmax(attn_score, axis=-1)
        is_nan = op.IsNaN(attn_weight)
        adj_attn_weight = op.Where(is_nan, 0.0, attn_weight)
        attn_weight = pattern.OrValue([adj_attn_weight, attn_weight])
        attn_output = op.MatMul(attn_weight, value)
        return attn_output

    def check(
        self,
        context,
        query: ir.Value | None,
        key: ir.Value | None,
        value: ir.Value | None,
        mask: ir.Value | None,
        key_format: str,
        **match_bindings,
    ):
        check_result = pattern.MatchResult()

        bindings: dict[str, Dim] = {}

        # Check that query/key/value have the expected shapes:
        # They all should have same batch-size (B) and number of heads (H). Conceptually, it is
        # different for Q and K/V, but the certain op implementations require them to be the same,
        # which is usually achieved via tiling/expanding K/V num-heads to match Q num-heads.
        # Query and Key should have same head-size (Dh) while value can have different head-size (Dv).
        # Key and Value should have same sequence length (Skv), while Query can have different sequence length (S).
        _fusion_utils.check_shape(bindings, query, ["B", "H", "S", "Dh"])
        if key_format == "BHSd":
            _fusion_utils.check_shape(bindings, key, ["B", "H", "Skv", "Dh"])
        else:
            assert key_format == "BSHd", f"Unexpected key format: {key_format}"
            _fusion_utils.check_shape(bindings, key, ["B", "Skv", "H", "Dh"])
        _fusion_utils.check_shape(bindings, value, ["B", "H", "Skv", "Dv"])

        def get_scale_value(tag_name: str, scale_name: str) -> float:
            scaling_type = match_bindings.get(tag_name, "None")
            if scaling_type == "None":
                return 1.0
            else:
                scale = match_bindings.get(scale_name)
                value = _ir_utils.get_singleton_value(scale)
                if value is None:
                    raise MatchFailureError(f"{scale_name} is not a scalar.", scale)
                if scaling_type == "Mul":
                    return value
                else:
                    assert scaling_type == "Div", f"Unexpected {scale_name} scaling operation"
                    return 1.0 / value

        query_scale_value = get_scale_value("query_scaling", "query_scale")
        key_scale_value = get_scale_value("key_scaling", "key_scale")
        qk_scale_value = get_scale_value("qk_scaling", "qk_scale")

        self._scale = query_scale_value * key_scale_value * qk_scale_value

        # If the scaling factor is the default one, we can skip passing it to SDPA.

        head_size = bindings["Dh"]
        if not isinstance(head_size, int):
            return check_result

        default_scaling_factor = 1.0 / math.sqrt(head_size)

        if math.isclose(self._scale, default_scaling_factor, rel_tol=1e-5, abs_tol=1e-8):
            # Pass no scaling factor to SDPA, SDPA will use the default scaling factor
            self._scale = None

        return check_result

    def rewrite(
        self,
        op,
        query: ir.Value | None,
        key: ir.Value | None,
        value: ir.Value | None,
        mask: ir.Value | None,
        key_format: str,
        **_,
    ):
        sdpa_args = [query, key, value]
        if mask is not None:
            sdpa_args.append(mask)
        # If the scale is None, SDPA will use the default scaling factor, which is 1/sqrt(head_size).
        return op.SDPA(
            *sdpa_args,
            scale=self._scale,
            key_format=key_format,
            _domain="ai.onnxruntime._fusion",
        )


# Dynamically create the rules
sdpa_rules = pattern.RewriteRuleSet(
    [
        SDPA.rule(),
    ]
)

fuse_sdpa = _fusion_utils.apply_fusion_rules(sdpa_rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import math
from typing import Callable, Sequence

import numpy as np

from onnxscript import ir, optimizer


def display_nodes(nodes: Sequence[ir.Node]) -> None:
    """Display a list of nodes in the order they appear in the graph."""
    if nodes:
        graph = nodes[0].graph
        if graph:
            # Display nodes in same order as in graph:
            # Currently doesn't handle (control-flow) subgraphs
            for node in graph:
                if node in nodes:
                    node.display()
        else:
            for node in nodes:
                node.display()


def display_slice(x: ir.Value | ir.Node, backward: bool = True, depth_limit: int = 5) -> None:
    """Display the (backward or forward) subgraph from a given value or node upto a certain depth."""
    slice = []

    def visit(node: ir.Node, depth):
        if node in slice:
            return
        slice.append(node)
        if depth < depth_limit:
            if backward:
                for inp in node.inputs:
                    if inp is not None and inp.producer() is not None:
                        visit(inp.producer(), depth + 1)  # type: ignore[arg-type]
            else:
                for out in node.outputs:
                    for consumer, _ in out.uses():
                        visit(consumer, depth + 1)

    if isinstance(x, ir.Node):
        visit(x, 0)
    elif isinstance(x, ir.Value) and x.producer() is not None:
        visit(x.producer(), 0)  # type: ignore[arg-type]
    display_nodes(slice)


def get_const_value(value: ir.Value) -> ir.TensorProtocol | None:
    node = value.producer()
    if node is not None:
        optimizer.basic_constant_propagation([node])
    return value.const_value


def get_numpy_value(val: ir.Value | None) -> np.ndarray | None:
    """Convenience wrapper to get (optional) numpy value from an optional IR Value.

    This is intended for use in optimizations/rewriting. Note that this does not
    yet handle the distinction between inputs with default values (values that are
    both graph inputs and graph initializers), which should not be treated as a
    constant, and true constant values. The caller should make the distinction, as
    a value does not contain enough information to determine this. (TODO)
    """
    if val is None:
        return None
    const_value = get_const_value(val)
    if const_value is not None:
        try:
            return const_value.numpy()
        except FileNotFoundError:
            # External data is not available.
            return None
    return None


def get_singleton_value(val: ir.Value | None, rank: int | Sequence[int] | None = None):
    """Returns element of a single element tensor constant value, and None otherwise.

    If an int rank is specified, it checks that the value has the given rank.
    If the rank is a sequence of ints, it checks that the value has one of the given ranks.

    Thus, `rank=0` checks for a scalar, `rank=1` checks for a 1D tensor, and
    `rank=(0,1)` checks for either a scalar or a 1D tensor.
    """
    np_val = get_numpy_value(val)
    if np_val is not None and np_val.size == 1:
        value = np_val.item()
        if (rank is None) or (isinstance(rank, int) and (np_val.ndim == rank)):
            return value
        if isinstance(rank, Sequence) and (np_val.ndim in rank):
            return value
    return None


def is_singleton_value(
    val: ir.Value | None,
    expected: float | int | Callable,
    *,
    rtol: float | None = None,
    rank: int | Sequence[int] | None = None,
) -> bool:
    """Returns True if the value is a single element tensor with given value, and False otherwise."""
    scalar = get_singleton_value(val, rank=rank)
    if scalar is None:
        return False
    if callable(expected):
        return expected(scalar)
    if isinstance(expected, int):
        return expected == scalar
    # rtol must be specified for float comparison
    assert rtol is not None
    return math.isclose(scalar, expected, rel_tol=rtol)


def is_1d_value(val: ir.Value | None, expected: list[int]) -> bool:
    """Returns True if the value is a 1d int64 tensor with given value, and False otherwise."""
    if val is None:
        return False
    if not isinstance(val.type, ir.TypeProtocol):
        return False
    np_val = get_numpy_value(val)
    if np_val is None:
        return False
    if (np_val.size != len(expected)) or (val.type.dtype != ir.DataType.INT64):
        return False
    values = np_val.tolist()
    return values == expected


def has_rank(value: ir.Value | None, rank: int) -> bool:
    """Returns True if the value is statically known to have the given rank, and False otherwise."""
    if value is None:
        return False
    shape = value.shape
    return (shape is not None) and (shape.rank() == rank)


def get_dim(value: ir.Value | None, dim: int) -> ir.SymbolicDim | int | None:
    """Returns the value of the given dimension, or None if it is not statically known."""
    if value is None:
        return None
    shape = value.shape
    if shape is None:
        return None
    if dim < 0:
        dim += shape.rank()
    if dim < 0 or dim >= shape.rank():
        return None
    return shape[dim]


def same_shape(shape1: ir.Shape | None, shape2: ir.Shape | None) -> bool:
    """Check if two shapes are semantically the same."""
    if shape1 is None or shape2 is None:
        return False

    # If any dim is unknown, the shapes are not the same
    if shape1.has_unknown_dim() or shape2.has_unknown_dim():
        return False

    return shape1 == shape2


def same_dim(dim1: ir.SymbolicDim | int, dim2: ir.SymbolicDim | int) -> bool:
    """Check if two dimensions are semantically the same."""
    if type(dim1) is not type(dim2):
        return False
    if isinstance(dim1, int) and isinstance(dim2, int):
        return dim1 == dim2
    assert isinstance(dim1, ir.SymbolicDim) and isinstance(dim2, ir.SymbolicDim)
    if dim1.value is None or dim2.value is None:
        return False
    return dim1.value == dim2.value



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# mypy: disable-error-code="misc,arg-type,type-arg,valid-type,assignment,return-value"
"""torch.ops.aten operators under the `sparse` module.

- No inplace operators.
- All functions should not have the script() decorator. This is because
    we want to delay the compilation of the function.
"""

from __future__ import annotations

from onnxscript.onnx_types import TensorType


def aten_sparse_sampled_addmm(
    self: TensorType, mat1: TensorType, mat2: TensorType, beta: float = 1.0, alpha: float = 1.0
) -> TensorType:
    """sparse_sampled_addmm(Tensor self, Tensor mat1, Tensor mat2, *, Scalar beta=1, Scalar alpha=1) -> Tensor"""

    raise NotImplementedError()
